"""
Comprehensive Offline Policy Evaluation (OPE) estimators.

This module provides a complete set of estimators for OPE.
"""

from abc import ABC, abstractmethod
from typing import Any, Callable, ClassVar, Dict, Optional, Tuple, Type

import numpy as np
from scipy.stats import bootstrap

from pybandits.base import Float01, PyBanditsBaseModel
from pybandits.pydantic_version_compatibility import (
    NonNegativeFloat,
    PositiveFloat,
    PositiveInt,
    PrivateAttr,
    validate_call,
)


class BaseOfflinePolicyEstimator(PyBanditsBaseModel, ABC):
    """Base class for all OPE estimators.

    This class defines the interface for all OPE estimators and provides a common method for estimating the policy value.

    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.
    """

    alpha: Float01 = 0.05
    n_bootstrap_samples: int = 10000
    random_state: Optional[int] = None
    name: ClassVar

    @classmethod
    def _check_array(
        cls,
        name: str,
        data: Dict[str, np.ndarray],
        ndim: PositiveInt,
        dtype: type,
        n_samples: PositiveInt,
        n_actions: Optional[PositiveInt] = None,
    ):
        if name in data:
            array = data[name]
            if array.ndim != ndim:
                raise ValueError(f"{name} must be a {ndim}D array.")
            if array.shape[0] != n_samples:
                raise ValueError(f"action and {name} must have the same length.")
            if array.dtype != dtype:
                raise ValueError(f"{name} must be a {dtype} array")
            if ndim > 1:
                if array.shape[1] != n_actions:
                    raise ValueError(f"{name} must have the same number of actions as the action array.")

    @classmethod
    def _check_sum(cls, name: str, data: Dict[str, np.ndarray]):
        if name in data:
            array = data[name]
            if not array.sum(axis=-1).all():
                raise ValueError(f"{name} must have at least one non-zero element on each column.")

    @classmethod
    def _check_inputs(cls, action: np.ndarray, **kwargs):
        """
        Check the inputs for the estimator.

        Parameters
        ----------
        action : np.ndarray
            Array of actions taken.
        """
        if action.ndim != 1:
            raise ValueError("action must be a 1D array.")
        if action.dtype != int:
            raise ValueError("action must be an integer array.")
        n_samples = action.shape[0]
        n_actions = np.unique(action).shape[0]

        for name, dtype in zip(["reward", "propensity_score", "expected_importance_weight"], [int, float, float]):
            cls._check_array(name, kwargs, 1, dtype, n_samples)

        for name in ["estimated_policy", "expected_reward"]:
            cls._check_array(name, kwargs, 2, float, n_samples, n_actions)

        for name in ["propensity_score", "estimated_policy", "expected_importance_weight"]:
            cls._check_sum(name, kwargs)

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def estimate_policy_value_with_confidence_interval(self, **kwargs) -> Tuple[float, float, float, float]:
        """
        Estimate the policy value with a confidence interval.

        Parameters
        ----------
        action : np.ndarray
            Array of actions taken.

        Returns
        -------
        Tuple[float, float, float, float]
            Estimated policy value, mean, lower bound, and upper bound of the confidence interval.
        """
        self._check_inputs(**kwargs)
        sample_reward = self.estimate_sample_rewards(**kwargs)
        estimated_policy_value = sample_reward.mean()
        bootstrap_result = bootstrap(
            data=(sample_reward,),
            statistic=np.mean,
            confidence_level=1 - self.alpha,
            n_resamples=self.n_bootstrap_samples,
            random_state=self.random_state,
        )
        low, high = bootstrap_result.confidence_interval
        std = bootstrap_result.standard_error
        return estimated_policy_value, low, high, std

    @abstractmethod
    def estimate_sample_rewards(self, **kwargs) -> np.ndarray:
        """
        Estimate sample rewards.

        Returns
        -------
        np.ndarray
            Estimated sample rewards.
        """
        pass


class ReplayMethod(BaseOfflinePolicyEstimator):
    """
    Replay Method estimator.

    This estimator is a simple baseline that estimates the policy value by averaging the rewards of the matched samples.

    References
    ----------
    Unbiased Offline Evaluation of Contextual-bandit-based News Article Recommendation Algorithms (Li, Chu, Langford, and Wang, 2011)
    https://arxiv.org/pdf/1003.5956

    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.

    """

    name = "rep"

    def estimate_sample_rewards(
        self, action: np.ndarray, reward: np.ndarray, estimated_policy: np.ndarray, **kwargs
    ) -> np.ndarray:
        """
        Estimate the sample rewards.

        Parameters
        ----------
        action : np.ndarray
            Array of actions taken.
        reward : np.ndarray
            Array of rewards corresponding to each action.
        estimated_policy : np.ndarray
            Array of action distributions.

        Returns
        -------
        sample_reward : np.ndarray
            Estimated sample rewards.
        """
        n_samples = action.shape[0]
        matched_evaluation_policy = estimated_policy[np.arange(n_samples), action]
        matched_action = matched_evaluation_policy == 1
        sample_reward = (
            reward * matched_action / matched_action.mean() if matched_action.any() else np.zeros_like(matched_action)
        )
        return sample_reward


class GeneralizedInverseProbabilityWeighting(BaseOfflinePolicyEstimator, ABC):
    """
    Abstract generalization of the Inverse Probability Weighting (IPW) estimator.

    References
    ----------
    Learning from Logged Implicit Exploration Data (Strehl, Langford, Li, and Kakade, 2010)
    https://arxiv.org/pdf/1003.0120

    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.
    """

    @abstractmethod
    def _get_importance_weights(self, **kwargs) -> np.ndarray:
        """
        Get the importance weights.

        Returns
        -------
        np.ndarray
            Array of importance weights.
        """
        pass

    def estimate_sample_rewards(self, reward: np.ndarray, shrinkage_method: Optional[Callable], **kwargs) -> np.ndarray:
        """
        Estimate the sample rewards.

        Parameters
        ----------
        reward : np.ndarray
            Array of rewards corresponding to each action.
        shrinkage_method : Optional[Callable]
            Shrinkage method for the importance weights.

        Returns
        -------
        sample_reward : np.ndarray
            Estimated sample rewards.
        """
        importance_weight = self._get_importance_weights(**kwargs)
        importance_weight = shrinkage_method(importance_weight) if shrinkage_method is not None else importance_weight
        sample_reward = reward * importance_weight
        return sample_reward


class InverseProbabilityWeighting(GeneralizedInverseProbabilityWeighting):
    """
    Inverse Probability Weighing (IPW) estimator.

    References
    ----------
    Learning from Logged Implicit Exploration Data (Strehl, Langford, Li, and Kakade, 2010)
    https://arxiv.org/pdf/1003.0120

    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.
    """

    name = "ipw"

    def estimate_sample_rewards(
        self,
        action: np.ndarray,
        reward: np.ndarray,
        propensity_score: np.ndarray,
        estimated_policy: np.ndarray,
        shrinkage_method: Optional[Callable] = None,
        **kwargs,
    ) -> np.ndarray:
        """
        Estimate the sample rewards.

        Parameters
        ----------
        action : np.ndarray
            Array of actions taken.
        reward : np.ndarray
            Array of rewards corresponding to each action.
        propensity_score : np.ndarray
            Array of propensity scores.
        estimated_policy : np.ndarray
            Array of action distributions.

        Returns
        -------
        sample_reward : np.ndarray
            Estimated sample rewards.
        """
        return super().estimate_sample_rewards(
            reward=reward,
            action=action,
            propensity_score=propensity_score,
            estimated_policy=estimated_policy,
            shrinkage_method=shrinkage_method,
        )

    def _get_importance_weights(
        self,
        action: np.ndarray,
        propensity_score: np.ndarray,
        estimated_policy: np.ndarray,
        **kwargs,
    ) -> np.ndarray:
        """
        Get the importance weights.

        Parameters
        ----------
        action : np.ndarray
            Array of actions taken
        propensity_score : np.ndarray
            Array of propensity scores.
        estimated_policy : np.ndarray
            Array of action distributions.

        Returns
        -------
        importance_weight : np.ndarray
            Array of importance weights.
        """
        n_samples = action.shape[0]
        importance_weight = estimated_policy[np.arange(n_samples), action] / propensity_score
        return importance_weight


class SelfNormalizedInverseProbabilityWeighting(InverseProbabilityWeighting):
    """
    Self-Normalized Inverse Propensity Score (SNIPS) estimator.

    References
    ----------
    The Self-normalized Estimator for Counterfactual Learning (Swaminathan and Joachims, 2015)
    https://papers.nips.cc/paper_files/paper/2015/file/39027dfad5138c9ca0c474d71db915c3-Paper.pdf

    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.
    """

    name = "snips"

    def estimate_sample_rewards(
        self,
        action: np.ndarray,
        reward: np.ndarray,
        propensity_score: np.ndarray,
        estimated_policy: np.ndarray,
        shrinkage_method: Optional[Callable] = None,
        **kwargs,
    ) -> np.ndarray:
        """
        Estimate the sample rewards.

        Parameters
        ----------
        action : np.ndarray
            Array of actions taken.
        reward : np.ndarray
            Array of rewards corresponding to each action.
        propensity_score : np.ndarray
            Array of propensity scores.
        estimated_policy : np.ndarray
            Array of action distributions.
        shrinkage_method : Optional[Callable]
            Shrinkage method for the importance weights.

        Returns
        -------
        sample_reward : np.ndarray
            Estimated sample rewards.
        """

        def self_normalized_shrink_weights(importance_weight: np.ndarray) -> np.ndarray:
            importance_weight = (
                shrinkage_method(importance_weight) if shrinkage_method is not None else importance_weight
            )
            return importance_weight / importance_weight.mean()

        sample_reward = super().estimate_sample_rewards(
            action=action,
            reward=reward,
            propensity_score=propensity_score,
            estimated_policy=estimated_policy,
            shrinkage_method=self_normalized_shrink_weights,
        )
        return sample_reward


class DirectMethod(BaseOfflinePolicyEstimator):
    """
    Direct Method (DM) estimator.

    This estimator uses the evaluation policy to Estimate the sample rewards.

    References
    ----------
    The Offset Tree for Learning with Partial Labels (Beygelzimer and Langford, 2009)
    https://arxiv.org/pdf/0812.4044


    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.
    """

    name = "dm"

    def estimate_sample_rewards(
        self,
        estimated_policy: np.ndarray,
        expected_reward: np.ndarray,
        **kwargs,
    ) -> np.ndarray:
        """
        Estimate the sample rewards.

        Parameters
        ----------
        estimated_policy : np.ndarray
            Array of action distributions.
        expected_reward : np.ndarray
            Array of expected rewards.

        Returns
        -------
        sample_reward : np.ndarray
            Estimated sample rewards.
        """
        n_samples = expected_reward.shape[0]
        base_expected_reward = expected_reward[np.arange(n_samples), :]
        evaluation_policy = estimated_policy[np.arange(n_samples), :]
        expected_reward = np.average(
            base_expected_reward,
            weights=evaluation_policy,
            axis=1,
        )
        return expected_reward


class GeneralizedDoublyRobust(BaseOfflinePolicyEstimator, ABC):
    """
    Abstract generalization of the Doubly Robust (DR) estimator.

    References
    ----------
    Doubly Robust Policy Evaluation and Optimization (Dudík, Erhan, Langford, and Li, 2014)
    https://arxiv.org/pdf/1503.02834

    More Robust Doubly Robust Off-policy Evaluation (Farajtabar, Chow, and Ghavamzadeh, 2018)
    https://arxiv.org/pdf/1802.03493


    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.
    """

    _alternative_method_cls: Type[InverseProbabilityWeighting]
    _dm: DirectMethod = PrivateAttr()
    _other_method: BaseOfflinePolicyEstimator = PrivateAttr()

    def model_post_init(self, __context: Any) -> None:
        self._dm = DirectMethod(
            alpha=self.alpha, n_bootstrap_samples=self.n_bootstrap_samples, random_state=self.random_state
        )
        self._other_method = self._alternative_method_cls(
            alpha=self.alpha, n_bootstrap_samples=self.n_bootstrap_samples, random_state=self.random_state
        )

    def estimate_sample_rewards(
        self,
        action: np.ndarray,
        reward: np.ndarray,
        propensity_score: np.ndarray,
        estimated_policy: np.ndarray,
        expected_reward: np.ndarray,
        **kwargs,
    ) -> np.ndarray:
        """
        Estimate the sample rewards.

        Parameters
        ----------
        action : np.ndarray
            Array of actions taken.
        reward : np.ndarray
            Array of rewards corresponding to each action.
        propensity_score : np.ndarray
            Array of propensity scores.
        estimated_policy : np.ndarray
            Array of action distributions.
        expected_reward : np.ndarray
            Array of expected rewards.

        Returns
        -------
        sample_reward : np.ndarray
            Estimated rewards.
        """
        dm_sample_reward = self._dm.estimate_sample_rewards(
            action=action, estimated_policy=estimated_policy, expected_reward=expected_reward
        )
        other_sample_reward = self._other_method.estimate_sample_rewards(
            action=action,
            reward=reward - dm_sample_reward,
            propensity_score=propensity_score,
            estimated_policy=estimated_policy,
            shrinkage_method=self._shrink_weights,
        )
        sample_reward = dm_sample_reward + other_sample_reward
        return sample_reward

    def _shrink_weights(self, importance_weight: np.ndarray) -> np.ndarray:
        return importance_weight


class DoublyRobust(GeneralizedDoublyRobust):
    """
     Doubly Robust (DR) estimator.

     References
     ----------
     Doubly Robust Policy Evaluation and Optimization (Dudík, Erhan, Langford, and Li, 2014)
     https://arxiv.org/pdf/1503.02834

    More Robust Doubly Robust Off-policy Evaluation (Farajtabar, Chow, and Ghavamzadeh, 2018)
    https://arxiv.org/pdf/1802.03493


     Parameters
     ----------
     alpha : Float01, default=0.05
         Significance level for confidence interval estimation.
     n_bootstrap_samples : int, default=10000
         Number of bootstrap samples for confidence interval estimation.
     random_state : int, default=None
         Random seed for bootstrap sampling.
    """

    name = "dr"
    _alternative_method_cls = InverseProbabilityWeighting


class SelfNormalizedDoublyRobust(GeneralizedDoublyRobust):
    """
    Self-Normalized Doubly Robust (SNDR) estimator.

    This estimator uses the self-normalized importance weights to combine the DR and IPS estimators.

    References
    ----------
    Intrinsically Efficient, Stable, and Bounded Off-Policy Evaluation for Reinforcement Learning (Kallus and Uehara, 2019)
    https://arxiv.org/pdf/1906.03735

    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.
    """

    name = "sndr"
    _alternative_method_cls = SelfNormalizedInverseProbabilityWeighting


class SwitchDoublyRobust(DoublyRobust):
    """
    Switch Doubly Robust (Switch-DR) estimator.

    This estimator uses a switching rule based on the propensity score to combine the DR and IPS estimators.

    References
    ----------
    Optimal and Adaptive Off-policy Evaluation in Contextual Bandits (Wang, Agarwal, and Dudik, 2017)
    https://arxiv.org/pdf/1507.02646

    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : Optional[int], default=None
        Random seed for bootstrap sampling.
    switch_threshold : float, default=inf
        Threshold for the importance weight to switch between the DR and IPS estimators.
    """

    name = "switch-dr"
    switch_threshold: float = float("inf")

    def _shrink_weights(self, importance_weight: np.ndarray) -> np.ndarray:
        switch_indicator = importance_weight >= self.switch_threshold
        return switch_indicator * importance_weight


class DoublyRobustWithOptimisticShrinkage(DoublyRobust):
    """
    Optimistic version of DRos estimator.

    This estimator uses a shrinkage factor to shrink the importance weight in the native DR.

    References
    ----------
    Doubly Robust Off-Policy Evaluation with Shrinkage (Su, Dimakopoulou, Krishnamurthy, and Dudik, 2020)
    https://arxiv.org/pdf/1907.09623

    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.
    shrinkage_factor : float, default=0.0
        Shrinkage factor for the importance weights.
        If set to 0 or infinity, the estimator is equivalent to the native DM or DR estimators, respectively.
    """

    shrinkage_factor: NonNegativeFloat = 0.0
    name = "dros-opt"

    def _shrink_weights(self, importance_weight: np.ndarray) -> np.ndarray:
        if self.shrinkage_factor == 0:
            return np.zeros_like(importance_weight)
        elif self.shrinkage_factor == float("inf"):
            return importance_weight
        return self.shrinkage_factor * importance_weight / (importance_weight**2 + self.shrinkage_factor)


class DoublyRobustWithPessimisticShrinkage(DoublyRobust):
    """
    Pessimistic version of DRos estimator.

    This estimator uses a shrinkage factor to shrink the importance weight in the native DR.

    References
    ----------
    Doubly Robust Off-Policy Evaluation with Shrinkage (Su, Dimakopoulou, Krishnamurthy, and Dudik, 2020)
    https://arxiv.org/pdf/1907.09623

    Parameters
    ----------
    alpha : Float01, default=0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, default=10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, default=None
        Random seed for bootstrap sampling.
    shrinkage_factor : float, default=0.0
        Shrinkage factor for the importance weights.
    """

    name = "dros-pess"
    shrinkage_factor: PositiveFloat = float("inf")

    def _shrink_weights(self, importance_weight: np.ndarray) -> np.ndarray:
        importance_weight = np.minimum(self.shrinkage_factor, importance_weight)
        return importance_weight


class SubGaussianInverseProbabilityWeighting(InverseProbabilityWeighting):
    """
    SubGaussian Inverse Probability Weighing estimator.

    References
    ----------
    Subgaussian and Differentiable Importance Sampling for Off-Policy Evaluation and Learning (Metelli, Russo, and Restelli, 2021)
    https://proceedings.neurips.cc/paper_files/paper/2021/file/4476b929e30dd0c4e8bdbcc82c6ba23a-Paper.pdf

    Parameters
    ----------
    alpha : Float01, defaults to 0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, defaults to 10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, defaults to None
        Random seed for bootstrap sampling.
    shrinkage_factor : Float01, defaults to 0.0
        Shrinkage factor for the importance weights.

    """

    name = "sg-ipw"
    shrinkage_factor: Float01 = 0.0

    def _shrink_weights(self, importance_weight: np.ndarray) -> np.ndarray:
        return importance_weight / (1 - self.shrinkage_factor + self.shrinkage_factor * importance_weight)


class SubGaussianDoublyRobust(GeneralizedDoublyRobust):
    """
    SubGaussian Doubly Robust estimator.

    References
    ----------
    Subgaussian and Differentiable Importance Sampling for Off-Policy Evaluation and Learning (Metelli, Russo, and Restelli, 2021)
    https://proceedings.neurips.cc/paper_files/paper/2021/file/4476b929e30dd0c4e8bdbcc82c6ba23a-Paper.pdf

    Parameters
    ----------
    alpha : Float01, defaults to 0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, defaults to 10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, defaults to None
        Random seed for bootstrap sampling.
    """

    name = "sg-dr"
    _alternative_method_cls = SubGaussianInverseProbabilityWeighting


class BalancedInverseProbabilityWeighting(GeneralizedInverseProbabilityWeighting):
    """
    Balanced Inverse Probability Weighing estimator.

    References
    ----------
    Balanced Off-Policy Evaluation in General Action Spaces (Sondhi, Arbour, and Dimmery, 2020)
    https://arxiv.org/pdf/1906.03694


    Parameters
    ----------
    alpha : Float01, defaults to 0.05
        Significance level for confidence interval estimation.
    n_bootstrap_samples : int, defaults to 10000
        Number of bootstrap samples for confidence interval estimation.
    random_state : int, defaults to None
        Random seed for bootstrap sampling.

    ----------
    Arjun Sondhi, David Arbour, and Drew Dimmery
    "Balanced Off-Policy Evaluation in General Action Spaces.", 2020.
    """

    name = "b-ipw"

    def _get_importance_weights(self, expected_importance_weight: np.ndarray, **kwargs) -> np.ndarray:
        """
        Get the importance weights.

        Parameters
        ----------
        expected_importance_weight : np.ndarray
            Array of expected importance weights.

        Returns
        -------
        expected_importance_weight : np.ndarray
            Array of expected importance weights.
        """
        return expected_importance_weight

    def estimate_sample_rewards(
        self,
        reward: np.ndarray,
        expected_importance_weight: np.ndarray,
        **kwargs,
    ) -> np.ndarray:
        """
        Estimate the sample rewards.

        Parameters
        ----------
        reward : np.ndarray
            Array of rewards corresponding to each action.
        expected_importance_weight : np.ndarray
            Array of expected importance weights.

        Returns
        -------
        sample_reward : np.ndarray
            Estimated rewards.
        """
        return super().estimate_sample_rewards(
            reward=reward, expected_importance_weight=expected_importance_weight, shrinkage_method=None
        )



import warnings
from abc import ABC, abstractmethod
from collections import defaultdict, deque
from typing import Any, Callable, ClassVar, Dict, Generic, List, Optional, Set, Tuple, Type, TypeVar, Union

import numpy as np
from numpy.typing import ArrayLike

from pybandits.base import (
    ACTION_IDS_PREFIX,
    QUANTITATIVE_ACTION_IDS_PREFIX,
    ActionId,
    BinaryReward,
    PositiveProbability,
    PyBanditsBaseModel,
)
from pybandits.base_model import BaseModel, BaseModelMO, BaseModelSO
from pybandits.model import (
    BaseBayesianNeuralNetwork,
    BaseBeta,
    BaseBetaMO,
    BayesianNeuralNetwork,
    BayesianNeuralNetworkCC,
    Beta,
    BetaCC,
    BetaMO,
    BetaMOCC,
    Model,
    ModelMO,
)
from pybandits.pydantic_version_compatibility import (
    PYDANTIC_VERSION_1,
    PYDANTIC_VERSION_2,
    Field,
    GenericModel,
    NonNegativeInt,
    NonPositiveInt,
    field_validator,
    pydantic_version,
    validate_call,
)
from pybandits.quantitative_model import (
    BaseCmabZoomingModel,
    BaseSmabZoomingModel,
    CmabZoomingModel,
    CmabZoomingModelCC,
    QuantitativeModel,
    SmabZoomingModel,
    SmabZoomingModelCC,
)
from pybandits.utils import classproperty, extract_argument_names_from_function


class ActionsManager(PyBanditsBaseModel, ABC):
    """
    Base class for managing actions and their associated models.
    The class allows to account for non-stationarity by providing an adaptive window scheme for action update.
    The change point detection is based on the adaptive windowing scheme.

    References
    ----------
    Scaling Multi-Armed Bandit Algorithms (Fouché et al., 2019)
    https://edouardfouche.com/publications/S-MAB_FOUCHE_KDD19.pdf

    Parameters
    ----------
    actions : Dict[ActionId, Model]
        The list of possible actions, and their associated Model.
    delta : Optional[PositiveProbability]
        The confidence level for the adaptive window. None for skipping the change point detection.
    """

    actions: Dict[ActionId, BaseModel]
    delta: Optional[PositiveProbability] = None
    _no_change_point: ClassVar[NonNegativeInt] = -1
    _min_adaptive_window_size: ClassVar[NonPositiveInt] = 10000
    _memory_parameters_suffix: ClassVar[str] = "_memory"
    actions_with_change: Set[Tuple[ActionId, NonNegativeInt]] = Field(default_factory=set)

    if pydantic_version == PYDANTIC_VERSION_1:

        class Config:
            arbitrary_types_allowed = True
            json_encoders = {deque: list}

    elif pydantic_version == PYDANTIC_VERSION_2:
        model_config = {"arbitrary_types_allowed": True, "json_encoders": {deque: list}}
    else:
        raise ValueError(f"Unsupported pydantic version: {pydantic_version}")

    @field_validator("actions", mode="after")
    @classmethod
    def at_least_one_action_is_defined(cls, v):
        # validate number of actions
        if len(v) == 0:
            raise AttributeError("At least one action should be defined.")
        elif len(v) == 1:
            warnings.warn("Only a single action was supplied. This MAB will be deterministic.")
        # validate that all actions are of the same configuration
        action_models = list(v.values())
        action_type = cls._get_field_type("actions")
        if any(not isinstance(action, action_type) for action in action_models):
            raise TypeError(f"All actions should follow {action_type} type.")
        return v

    @classmethod
    def _get_expected_memory_length(cls, actions: Dict[ActionId, BaseModel]) -> NonNegativeInt:
        """
        Get the expected memory length for the adaptive window.

        Parameters
        ----------
        actions : Dict[ActionId, BaseModel]
            The list of possible actions, and their associated Model.

        Returns
        -------
        NonNegativeInt
            The expected memory length.
        """
        if not actions:
            raise AttributeError("At least one action should be defined.")
        reference_model = list(actions.values())[0]
        if isinstance(reference_model, BaseModelSO):
            expected_memory_length_for_inf = sum([action_model.count - 2 for action_model in actions.values()])
        elif isinstance(reference_model, BaseModelMO):
            expected_memory_length_for_inf = sum(
                [action_model.models[0].count - 2 for action_model in actions.values()]
            )
        else:
            raise ValueError(f"Model type {type(reference_model)} not supported.")
        return expected_memory_length_for_inf

    def __init__(
        self,
        delta: Optional[PositiveProbability] = None,
        actions: Optional[Dict[ActionId, Model]] = None,
        action_ids: Optional[Set[ActionId]] = None,
        quantitative_action_ids: Optional[Set[ActionId]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        actions_with_change: Optional[Set[Tuple[ActionId, NonNegativeInt]]] = None,
    ):
        kwargs = kwargs or {}
        actions = self._instantiate_actions(
            actions=actions, action_ids=action_ids, quantitative_action_ids=quantitative_action_ids, kwargs=kwargs
        )
        actions_with_change = actions_with_change or set()
        super().__init__(actions=actions, delta=delta, actions_with_change=actions_with_change)

    def _validate_update_params(
        self,
        actions: List[ActionId],
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]],
        quantities: Optional[List[Union[float, List[float], None]]] = None,
        **kwargs,
    ):
        """
        Verify that the given list of action IDs is a subset of the currently defined actions and that
         the rewards type matches the strategy type.

        Parameters
        ----------
        actions : List[ActionId]
            The selected action for each sample.
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]]
            The reward for each sample.
        quantities : Optional[List[Union[float, List[float], None]]]
            The value associated with each action. If none, the value is not used, i.e. non-quantitative action.
        """
        invalid = set(actions) - set(self.actions.keys())
        if invalid:
            raise AttributeError(f"The following invalid action(s) were specified: {invalid}.")
        self._validate_params_lengths(actions=actions, rewards=rewards, quantities=quantities, **kwargs)
        if quantities is None:
            if not all(isinstance(self.actions[action], (Model, ModelMO)) for action in actions):
                raise ValueError("Quantitative actions require defined quantities.")
        else:
            if not all(
                q is not None for a, q in zip(actions, quantities) if isinstance(self.actions[a], QuantitativeModel)
            ):
                raise ValueError("Quantitative actions require defined quantities.")
            if not all(q is None for a, q in zip(actions, quantities) if isinstance(self.actions[a], (Model, ModelMO))):
                raise ValueError("regular actions should not have defined quantities.")

    @classmethod
    def _to_memory_key(cls, key: str) -> str:
        return f"{key}{cls._memory_parameters_suffix}"

    @classmethod
    def _to_key(cls, key: str) -> str:
        return key.replace(cls._memory_parameters_suffix, "")

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def update(
        self,
        actions: List[ActionId],
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]],
        quantities: Optional[List[Union[float, List[float], None]]] = None,
        actions_memory: Optional[List[ActionId]] = None,
        rewards_memory: Optional[Union[List[BinaryReward], List[List[BinaryReward]]]] = None,
        **kwargs,
    ):
        """
        Update the models associated with the given actions using the provided rewards.
        For adaptive window size, the update by resetting the action models and retraining them on the new data.

        Parameters
        ----------
        actions : List[ActionId]
            The selected action for each sample.
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]]
            The reward for each sample.
        quantities : Optional[List[Union[float, List[float], None]]]
            The value associated with each action. If none, the value is not used, i.e. non-quantitative action.
        actions_memory : Optional[List[ActionId]]
            List of previously selected actions.
        rewards_memory : Optional[Union[List[BinaryReward], List[List[BinaryReward]]]]
            List of previously collected rewards.
        """
        self.actions_with_change.clear()
        if self.delta is None and (actions_memory or rewards_memory):
            raise AttributeError("Adaptive window size is not set, so memory should not be provided.")
        if self.delta is not None and (actions_memory is None or rewards_memory is None):
            warnings.warn("Adaptive window size is set, but memory was not provided.")
            actions_memory = []
            rewards_memory = []

        update_kwargs = {k: v for k, v in kwargs.items() if not k.endswith(self._memory_parameters_suffix)}
        memory_kwargs = {k: v for k, v in kwargs.items() if k.endswith(self._memory_parameters_suffix)}
        self._validate_update_params(actions, rewards, quantities, **update_kwargs)
        self._validate_params_lengths(actions_memory=actions_memory, rewards_memory=rewards_memory, **memory_kwargs)
        update_keys = tuple(update_kwargs.keys())

        if actions_memory is not None:
            actions_memory, rewards_memory, memory_kwargs = self._maybe_trim_memory(
                actions_memory, rewards_memory, memory_kwargs
            )
            residual_memory_len = len(actions_memory)
            if residual_memory_len < self._min_adaptive_window_size:
                warnings.warn("The adaptive window size too small value. Consider increasing it for better results.")
            actions_memory.extend(actions)
            rewards_memory.extend(rewards)
            for key in update_keys:
                memory_key = self._to_memory_key(key)
                if isinstance(update_kwargs[key], list):
                    memory_kwargs[memory_key].extend(update_kwargs[key])
                elif isinstance(update_kwargs[key], np.ndarray):
                    memory_kwargs[memory_key] = (
                        np.concatenate(
                            (memory_kwargs[f"{key}{self._memory_parameters_suffix}"], update_kwargs[key]), axis=0
                        )
                        if memory_kwargs[f"{key}{self._memory_parameters_suffix}"] is not None
                        else update_kwargs[key]
                    )

            if (
                last_change_point := self._get_last_change_point(residual_memory_len, actions_memory, rewards_memory)
            ) != self._no_change_point:
                actions_memory, rewards_memory, memory_kwargs = self._slice_memory(
                    len(actions_memory) - last_change_point, actions_memory, rewards_memory, memory_kwargs
                )

                for action_model in self.actions.values():
                    if not isinstance(action_model, QuantitativeModel):
                        action_model.reset()
                regular_actions = [a for a in actions if not isinstance(self.actions[a], QuantitativeModel)]
                quantitative_actions = [a for a in actions if isinstance(self.actions[a], QuantitativeModel)]

                if regular_actions:
                    regular_rewards = [
                        r
                        for a, r in zip(actions_memory, rewards_memory)
                        if not isinstance(self.actions[a], QuantitativeModel)
                    ]
                    filtered_update_kwargs = {
                        self._to_key(k): [
                            v
                            for a, v in zip(actions_memory, values)
                            if not isinstance(self.actions[a], QuantitativeModel)
                        ]
                        if isinstance(values, list)
                        else values
                        for k, values in memory_kwargs.items()
                    }
                    self._update_actions(regular_actions, regular_rewards, None, **filtered_update_kwargs)

                if quantitative_actions:
                    filtered_quantitative_kwargs = {
                        k: [v for a, v in zip(actions, values) if isinstance(self.actions[a], QuantitativeModel)]
                        if isinstance(values, list)
                        else values
                        for k, values in update_kwargs.items()
                    }
                    quantitative_rewards = [
                        r for a, r in zip(actions, rewards) if isinstance(self.actions[a], QuantitativeModel)
                    ]
                    quantitative_quantities = [
                        q for a, q in zip(actions, quantities) if isinstance(self.actions[a], QuantitativeModel)
                    ]
                    self._update_actions(
                        quantitative_actions,
                        quantitative_rewards,
                        quantitative_quantities,
                        **filtered_quantitative_kwargs,
                    )
            else:
                self._update_actions(actions, rewards, quantities, **update_kwargs)
        else:
            self._update_actions(actions, rewards, quantities, **update_kwargs)

    @staticmethod
    def _slice_memory(
        memory_len: NonNegativeInt,
        actions_memory: List[ActionId],
        rewards_memory: List[BinaryReward],
        memory_kwargs: Dict[str, Any],
    ) -> Tuple[List[ActionId], List[BinaryReward], Dict[str, Any]]:
        """
        Slice all memory parameters to memory_len length.

        Parameters
        ----------
        memory_len : NonNegativeInt
            Expected memory length after the slicing.
        actions_memory : List[ActionId]
            List of previously selected actions.
        rewards_memory : List[BinaryReward]
            List of previously collected rewards.
        memory_kwargs : Dict[str, Any]
            The memory kwargs.

        Returns
        -------
        actions_memory : List[ActionId]
            List of previously selected actions with maximum length of memory_len.
        rewards_memory : List[BinaryReward]
            List of previously collected rewards with maximum length of memory_len.
        memory_kwargs : Dict[str, Any]
            The memory kwargs with values of maximum length of memory_len.
        """
        if len(actions_memory) > memory_len:
            actions_memory = actions_memory[-memory_len:]
            rewards_memory = rewards_memory[-memory_len:]
            for memory_key, memory_value in memory_kwargs.items():
                if memory_value is not None:
                    memory_kwargs[memory_key] = memory_value[-memory_len:]
        return actions_memory, rewards_memory, memory_kwargs

    def _maybe_trim_memory(
        self,
        actions_memory: List[ActionId],
        rewards_memory: Union[List[BinaryReward], List[List[BinaryReward]]],
        memory_kwargs: Dict[str, Any],
    ) -> Tuple[List[ActionId], List[BinaryReward], Dict[str, Any]]:
        """
        Trim the memory to the adaptive window size.

        Parameters
        ----------
        actions_memory : List[ActionId]
            List of previously selected actions.
        rewards_memory : Union[List[BinaryReward], List[List[BinaryReward]]]
            List of previously collected rewards.
        memory_kwargs : Dict[str, Any]
            The memory kwargs.

        Returns
        -------
        actions_memory : List[ActionId]
            List of previously selected actions with maximum length of memory_len.
        rewards_memory : List[BinaryReward]
            List of previously collected rewards with maximum length of memory_len.
        memory_kwargs : Dict[str, Any]
            The memory kwargs with values of maximum length of memory_len.
        """
        action_stats = self._action_stats
        maximum_memory_length = self._get_memory_len_from_action_stats(action_stats)
        if len(actions_memory) > maximum_memory_length:
            warnings.warn(f"Input memory is longer then expected. Leaving only last {maximum_memory_length} elements.")
            actions_memory, rewards_memory, memory_kwargs = self._slice_memory(
                maximum_memory_length, actions_memory, rewards_memory, memory_kwargs
            )
        for action_id, (expected_successes, expected_trials) in action_stats.items():
            actual_trials = np.sum([1 for a in actions_memory if a == action_id])
            actual_successes = np.sum(
                np.array([r for a, r in zip(actions_memory, rewards_memory) if a == action_id]).reshape(
                    (-1, expected_successes.shape[1])
                ),
                axis=0,
                keepdims=True,
            )

            if np.any(actual_trials > expected_trials):
                raise ValueError(f"Memory for action {action_id} is larger than expected.")
            elif actual_trials == expected_trials[0][0]:
                if not np.array_equal(actual_successes, expected_successes):
                    raise ValueError(f"Memory for action {action_id} is not consistent with the expected stats.")
            else:
                if np.any(actual_successes > expected_successes):
                    raise ValueError(f"Memory for action {action_id} is not consistent with the expected stats.")

        return actions_memory, rewards_memory, memory_kwargs

    def _get_memory_len_from_action_stats(
        self, action_stats: Dict[ActionId, Tuple[ArrayLike, ArrayLike]]
    ) -> NonNegativeInt:
        """
        Calculate total memory length from action statistics.

        Parameters
        ----------
        action_stats : Dict[ActionId, Tuple[ArrayLike, ArrayLike]]
            Dictionary mapping action IDs to tuples of (successes, trials) arrays.

        Returns
        -------
        NonNegativeInt
            Total number of trials across all actions.
        """

        return sum([v[1][0][0] for v in action_stats.values()])

    @property
    def _action_stats(self) -> Dict[ActionId, Tuple[np.ndarray, np.ndarray]]:
        """
        Get current statistics for all actions.

        Returns
        -------
        action_stats : Dict[ActionId, Tuple[np.ndarray, np.ndarray]]
            Dictionary mapping action IDs to tuples of (successes, trials) arrays.
        """
        action_stats = {action_id: self._extract_current_stats_for_action(action_id) for action_id in self.actions}
        return action_stats

    @property
    def maximum_memory_length(self) -> NonNegativeInt:
        """
        Get maximum possible memory length based on current action statistics.

        Returns
        -------
        NonNegativeInt
            Maximum memory length allowed.
        """
        return self._get_memory_len_from_action_stats(self._action_stats)

    @abstractmethod
    def _update_actions(
        self,
        actions: List[ActionId],
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]],
        quantities: Optional[List[Union[float, List[float], None]]],
        **kwargs,
    ):
        """
        Update the models associated with the given actions using the provided rewards.

        Parameters
        ----------
        actions : List[ActionId]
            The selected action for each sample.
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]]
            The reward for each sample.
        quantities : Optional[List[Union[float, List[float], None]]]
            The value associated with each action. If none, the value is not used, i.e. non-quantitative action.
        """

    def _get_last_change_point(
        self,
        residual_memory_len: NonNegativeInt,
        actions_memory: List[ActionId],
        rewards_memory: Union[List[BinaryReward], List[List[BinaryReward]]],
    ) -> NonNegativeInt:
        """
        Get the last change point among all actions.

        Parameters
        ----------
        residual_memory_len : NonNegativeInt
            The length of the residual memory.
        actions_memory : List[ActionId]
            List of previously selected actions.
        rewards_memory : List[BinaryReward]
            List of previously collected rewards.

        Returns
        -------
        NonNegativeInt
            The last change point. 0 if no change point is found.
        """
        last_change_point = self._no_change_point
        for action_id, action_model in self.actions.items():
            if not isinstance(action_model, QuantitativeModel):
                change_point = self._get_last_change_point_for_action(
                    action_id=action_id,
                    residual_memory_len=residual_memory_len,
                    actions_memory=actions_memory,
                    rewards_memory=rewards_memory,
                )
                if change_point != self._no_change_point:
                    self.actions_with_change.add((action_id, change_point))
                    last_change_point = max(last_change_point, change_point)
        return last_change_point

    def _get_threshold(self, past_trials: np.ndarray, present_trials: np.ndarray) -> np.ndarray:
        """
        Get the threshold for the given past window and present window.

        Parameters
        ----------
        past_trials : np.ndarray
            The number of trials in the past window.
        present_trials : np.ndarray
            The number of trials in the present window.

        Returns
        -------
        threshold : np.ndarray
            The threshold value.
        """
        full_trials = past_trials + present_trials
        harmonic_sum = 1 / past_trials + 1 / present_trials
        threshold = past_trials * present_trials * np.sqrt((harmonic_sum / 2) * np.log(4 * full_trials / self.delta))
        return threshold

    def _get_last_change_point_for_action(
        self,
        action_id: ActionId,
        residual_memory_len: NonNegativeInt,
        actions_memory: List[ActionId],
        rewards_memory: Union[List[BinaryReward], List[List[BinaryReward]]],
    ) -> int:
        """
        Get the last change point for the given action.

        Parameters
        ----------
        action_id : ActionId
            The action ID.
        actions_memory : List[ActionId]
            List of previously selected actions.
        rewards_memory : List[BinaryReward]
            List of previously collected rewards.

        Returns
        -------
        NonNegativeInt
            The last change point for the given action. -1 if no change point is found.
        """
        action_index = np.nonzero([a == action_id for a in actions_memory])[0].tolist()

        rewards_window = [rewards_memory[i] for i in action_index]
        window_length = len(rewards_window)
        if window_length < 2:
            return self._no_change_point
        cumulative_reward = np.cumsum(np.array(rewards_window), axis=0)
        if cumulative_reward.ndim == 1:
            cumulative_reward = cumulative_reward[:, np.newaxis]

        current_sum, current_trials = self._extract_current_stats_for_action(action_id)

        # n_successes and n_failures already take into account the statistics of remaining elements from last
        # memory update, so their statistics are removed for consistency.
        if residual_memory_len:
            projected_residual_memory_len = len([index for index in action_index if index < residual_memory_len])
            current_sum -= cumulative_reward[projected_residual_memory_len - 1]
            current_trials -= projected_residual_memory_len
        initial_start_index = 0 if np.sum(current_trials) else 1

        base_range = np.arange(initial_start_index, window_length).reshape(-1, 1)
        past_sums = np.concatenate((current_sum, current_sum + cumulative_reward[:-1]))
        present_sums = cumulative_reward[-1] - np.concatenate(
            (np.zeros((1, cumulative_reward.shape[-1])), cumulative_reward[:-1])
        )
        start_index = initial_start_index
        while start_index < window_length:
            if initial_start_index == 0 and start_index == 1:  # After first iteration, dismiss the current memory
                past_sums -= current_sum
                current_trials = np.zeros_like(current_trials)

            relevant_range = base_range[(start_index - initial_start_index) :]

            past_trials = relevant_range + current_trials
            present_trials = window_length - relevant_range

            thresholds = self._get_threshold(past_trials, present_trials)
            change_points = np.nonzero(
                np.any(
                    np.abs(past_sums[start_index:] * present_trials - present_sums[start_index:] * past_trials)
                    > thresholds,
                    axis=1,
                )
            )[0]

            if not change_points.size:
                break
            start_index += 1

        if start_index == initial_start_index:
            return self._no_change_point

        return action_index[min(start_index, window_length - 1)]

    def _extract_current_stats_for_action(self, action_id: ActionId) -> Tuple[np.ndarray, np.ndarray]:
        """
        Extract the current statistics for the given action.
        The statistics include the number of successes and the number of trials for each action.
        Since `n_successes` and `n_failures` are initialized as 1 in Model class,
        we reduce 1 from `n_successes` to get the actual number of successes.
        Similarly, we reduce 2 from `count` to get the actual number of trials.

        Parameters
        ----------
        action_id : ActionId
            The action ID.

        Returns
        -------
        current_sum : np.ndarray
            The number of successes for the given action for each of the objectives.
        current_trials : np.ndarray
            The number of trials for the given action for each of the objectives.
        """
        action_model = self.actions[action_id]
        if isinstance(action_model, BaseModelSO):
            current_sum = np.array([action_model.n_successes - 1]).reshape((1, -1))
            current_trials = np.array([action_model.count - 2]).reshape((1, -1))

        elif isinstance(action_model, BaseModelMO):
            current_sum = np.array([model.n_successes - 1 for model in action_model.models]).reshape((1, -1))
            current_trials = np.array([model.count - 2 for model in action_model.models]).reshape((1, -1))
        else:
            raise TypeError(f"Model type {type(action_model)} not supported.")
        return current_sum, current_trials

    @classmethod
    def _instantiate_actions(
        cls,
        actions: Optional[Dict[ActionId, Model]],
        action_ids: Optional[Set[ActionId]],
        quantitative_action_ids: Optional[Set[ActionId]],
        kwargs: Dict[str, Any],
    ):
        """
        Utility function to instantiate the action models based on the provided kwargs.

        Parameters
        ----------
        actions : Optional[Dict[ActionId, Model]]
            The list of possible actions and their associated models.
        action_ids : Optional[Set[ActionId]]
            The list of possible actions.
        quantitative_action_ids : Optional[Set[ActionId]]
            The list of quantitative actions.
        kwargs : Dict[str, Any]
            Additional parameters for the mab and for the action model.

        Returns
        -------
        actions : Dict[ActionId, Model]
            Dictionary of actions and the parameters of their associated model.
        """
        if actions is None:
            action_specific_kwargs, quantitative_action_specific_kwargs = cls._extract_action_specific_kwargs(kwargs)

            # Extract inner_action_ids
            inner_action_ids = action_ids or set(action_specific_kwargs)
            inner_quantitative_action_ids = quantitative_action_ids or set(quantitative_action_specific_kwargs)
            if not inner_action_ids and not inner_quantitative_action_ids:
                raise ValueError("At least one action should be defined.")

            # Assign model for each action
            (
                model_cold_start,
                quantitative_model_cold_start,
                action_general_kwargs,
                quantitative_action_general_kwargs,
            ) = cls._extract_action_model_class_and_attributes(kwargs)

            # Instantiate the actions
            actions = {}
            for action_ids, cold_start, general_kwargs, specific_kwargs in zip(
                [inner_action_ids, inner_quantitative_action_ids],
                [model_cold_start, quantitative_model_cold_start],
                [action_general_kwargs, quantitative_action_general_kwargs],
                [action_specific_kwargs, quantitative_action_specific_kwargs],
            ):
                for a in action_ids:
                    actions[a] = cold_start(**general_kwargs, **specific_kwargs.get(a, {}))

        return actions

    @staticmethod
    def _extract_action_specific_kwargs(kwargs: Dict[str, Any]) -> Tuple[Dict[str, Dict], Dict[str, Dict]]:
        """
        Utility function to extract kwargs that are specific for each action when constructing the action model.

        Parameters
        ----------
        kwargs : Dict[str, Any]
            Additional parameters for the mab and for the action model.

        Returns
        -------
        action_specific_kwargs : Dict[str, Dict]
            Dictionary of actions and the parameters of their associated model.
        quantitative_action_specific_kwargs : Dict[str, Dict]
            Dictionary of quantitative actions and the parameters of their associated model.
        kwargs : Dict[str, Any]
            Dictionary of parameters and their quantities, without the action_specific_kwargs.
        """
        action_specific_kwargs = defaultdict(dict)
        quantitative_action_specific_kwargs = defaultdict(dict)
        for keyword in list(kwargs):
            argument = kwargs[keyword]
            for prefix, target_kwargs in zip(
                [ACTION_IDS_PREFIX, QUANTITATIVE_ACTION_IDS_PREFIX],
                [action_specific_kwargs, quantitative_action_specific_kwargs],
            ):
                if keyword.startswith(prefix) and type(argument) is dict:
                    kwargs.pop(keyword)
                    inner_keyword = keyword.split(prefix)[1]
                    for action_id, value in argument.items():
                        target_kwargs[action_id][inner_keyword] = value
        return dict(action_specific_kwargs), dict(quantitative_action_specific_kwargs)

    @classmethod
    def _extract_action_model_class_and_attributes(
        cls, kwargs
    ) -> Tuple[Callable, Callable, Dict[str, Dict], Dict[str, Dict]]:
        """
        Utility function to extract kwargs that are specific for each action when constructing the action model.

        Parameters
        ----------
        kwargs : Dict[str, Any]
            Additional parameters for the mab and for the action model.

        Returns
        -------
        action_model_cold_start : Callable
            Function handle for factoring the required action model.
        quantitative_action_model_cold_start : Callable
            Function handle for factoring the required quantitative action model.
        action_general_kwargs : Dict[str, any]
            Dictionary of parameters and their values for the action model.
        quantitative_action_general_kwargs : Dict[str, any]
            Dictionary of parameters and their values for the quantitative action model.
        """
        action_model_classes = cls._action_model_classes
        if len(action_model_classes) > 2:
            raise ValueError("Only up to two types of action models are supported.")
        quantitative_model_cold_start = model_cold_start = lambda **kwargs: None  # dummy callable
        action_general_kwargs = quantitative_action_general_kwargs = None
        for action_model_class in action_model_classes:
            if hasattr(action_model_class, "cold_start"):
                action_model_cold_start = action_model_class.cold_start
                action_model_attributes = extract_argument_names_from_function(action_model_cold_start)
                # cover for cold_start kwargs
                action_model_attributes = action_model_attributes + extract_argument_names_from_function(
                    action_model_class
                )
            else:
                action_model_cold_start = action_model_class
                action_model_attributes = extract_argument_names_from_function(action_model_cold_start)
            general_kwargs = {k: kwargs.pop(k) for k in action_model_attributes if k in kwargs.keys()}

            if issubclass(action_model_class, (Model, ModelMO)):
                model_cold_start = action_model_cold_start
                action_general_kwargs = general_kwargs
            elif issubclass(action_model_class, QuantitativeModel):
                quantitative_model_cold_start = action_model_cold_start
                quantitative_action_general_kwargs = general_kwargs
            else:
                raise TypeError(f"Unsupported action model class: {action_model_class}")

        return (
            model_cold_start,
            quantitative_model_cold_start,
            action_general_kwargs,
            quantitative_action_general_kwargs,
        )

    @classproperty
    def _action_model_classes(cls) -> Tuple[Type[BaseModel], ...]:
        """
        Utility function to extract the action model classes from the actions field.

        Returns
        -------
        action_model_classes : Tuple[Type[BaseModel],...]
            Tuple of action model classes.
        """
        action_model_type = cls._get_field_type("actions")
        action_model_classes = action_model_type if isinstance(action_model_type, tuple) else (action_model_type,)
        return action_model_classes


SmabModelType = TypeVar("SmabModelType", bound=Union[BaseBeta, BaseBetaMO, BaseSmabZoomingModel])


class SmabActionsManager(ActionsManager, GenericModel, Generic[SmabModelType]):
    """
    Manages actions and their associated models for sMAB models.
    The class allows to account for non-stationarity by providing an adaptive window scheme for action update.

    Parameters
    ----------
    actions : Dict[ActionId, BaseBeta]
        The list of possible actions, and their associated Model.
    delta : Optional[PositiveProbability], 0.1 if not specified.
        The confidence level for the adaptive window.
    """

    actions: Dict[ActionId, SmabModelType]

    @field_validator("actions", mode="after")
    @classmethod
    def all_actions_have_same_number_of_objectives(cls, actions: Dict[ActionId, SmabModelType]):
        n_objs_per_action = [len(beta.models) if hasattr(beta, "models") else None for beta in actions.values()]
        if len(set(n_objs_per_action)) != 1:
            raise ValueError("All actions should have the same number of objectives")
        return actions

    @validate_call
    def update(
        self,
        actions: List[ActionId],
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]],
        quantities: Optional[List[Union[float, List[float], None]]],
        actions_memory: Optional[List[ActionId]] = None,
        rewards_memory: Optional[Union[List[BinaryReward], List[List[BinaryReward]]]] = None,
    ):
        """
        Update the models associated with the given actions using the provided rewards.
        For adaptive window size, the update by resetting the action models and retraining them on the new data.

        Parameters
        ----------
        actions : List[ActionId]
            The selected action for each sample.
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]]
            The reward for each sample.
        quantities : Optional[List[Union[float, List[float], None]]]
            The value associated with each action. If none, the value is not used, i.e. non-quantitative action.
        actions_memory : Optional[List[ActionId]]
            List of previously selected actions.
        rewards_memory : Optional[Union[List[BinaryReward], List[List[BinaryReward]]]]
            List of previously collected rewards.
        """
        super().update(actions, rewards, quantities, actions_memory, rewards_memory)

    def _update_actions(
        self,
        actions: List[ActionId],
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]],
        quantities: Optional[List[Union[float, List[float], None]]],
    ):
        """
        Update the stochastic Bernoulli bandit given the list of selected actions and their corresponding binary
        rewards.

        Parameters
        ----------
        actions : List[ActionId] of shape (n_samples,), e.g. ['a1', 'a2', 'a3', 'a4', 'a5']
            The selected action for each sample.
        rewards : Union[List[BinaryReward], List[List[BinaryReward]]],
            if nested list, len() should follow shape of (n_samples, n_objectives)
            The binary reward for each sample.
                If strategy is not MultiObjectiveBandit, rewards should be a list, e.g.
                    rewards = [1, 0, 1, 1, 1, ...]
                If strategy is MultiObjectiveBandit, rewards should be a list of list, e.g. (with n_objectives=2):
                    rewards = [[1, 1], [1, 0], [1, 1], [1, 0], [1, 1], ...]
        quantities : Optional[List[Union[float, List[float], None]]]
            The value associated with each action. If none, the value is not used, i.e. non-quantitative action.
        """

        rewards_dict = defaultdict(list)

        if quantities is None:
            for a, r in zip(actions, rewards):
                rewards_dict[a].append(r)
            for a in set(actions):
                self.actions[a].update(rewards=rewards_dict[a])
        else:
            quantities = quantities[-len(actions) :]
            quantities_dict = defaultdict(list)
            for a, v, r in zip(actions, quantities, rewards):
                if v is not None:
                    quantities_dict[a].append(v)
                rewards_dict[a].append(r)
            for a in set(actions):
                if quantities_dict[a]:  # quantitative action
                    self.actions[a].update(rewards=rewards_dict[a], quantities=quantities_dict[a])
                else:  # non-quantitative action
                    self.actions[a].update(rewards=rewards_dict[a])


CmabModelType = TypeVar("CmabModelType", bound=Union[BaseBayesianNeuralNetwork, BaseCmabZoomingModel])


class CmabActionsManager(ActionsManager, GenericModel, Generic[CmabModelType]):
    """
    Manages actions and their associated models for cMAB models.
    The class allows to account for non-stationarity by providing an adaptive window scheme for action update.

    Parameters
    ----------
    actions : Dict[ActionId, BaseBayesianNeuralNetwork]
        The list of possible actions, and their associated Model.
    delta : Optional[PositiveProbability], 0.1 if not specified.
        The confidence level for the adaptive window.
    """

    actions: Dict[ActionId, CmabModelType]

    @staticmethod
    def _maybe_crawl_model(model: Union[BaseBayesianNeuralNetwork, BaseCmabZoomingModel]):
        return list(model.sub_actions.values())[0] if isinstance(model, BaseCmabZoomingModel) else model

    @field_validator("actions", mode="after")
    @classmethod
    def check_models(cls, v):
        action_models = list(v.values())
        first_action = action_models[0]
        test_first_action = cls._maybe_crawl_model(first_action)
        for action in action_models[1:]:
            test_action = cls._maybe_crawl_model(action)
            if not test_first_action.input_dim == test_action.input_dim:
                raise AttributeError("All actions should have the same input size.")
            if not test_first_action.update_method == test_action.update_method:
                raise AttributeError("All actions should have the same update method.")
            if not test_first_action.update_kwargs == test_action.update_kwargs:
                raise AttributeError("All actions should have the same update kwargs.")
        return v

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def update(
        self,
        actions: List[ActionId],
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]],
        quantities: Optional[List[Union[float, List[float], None]]],
        context: ArrayLike,
        actions_memory: Optional[List[ActionId]] = None,
        rewards_memory: Optional[Union[List[BinaryReward], List[List[BinaryReward]]]] = None,
        context_memory: Optional[ArrayLike] = None,
    ):
        """
        Update the models associated with the given actions using the provided rewards.
        For adaptive window size, the update by resetting the action models and retraining them on the new data.

        Parameters
        ----------
        actions : List[ActionId]
            The selected action for each sample.
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]]
            The reward for each sample.
        quantities : Optional[List[Union[float, List[float], None]]]
            The value associated with each action. If none, the value is not used, i.e. non-quantitative action.
        context: ArrayLike of shape (n_samples, n_features)
            Matrix of contextual features.
        actions_memory : Optional[List[ActionId]]
            List of previously selected actions.
        rewards_memory : Optional[Union[List[BinaryReward], List[List[BinaryReward]]]]
            List of previously collected rewards.
        context_memory : Optional[ArrayLike] of shape (n_samples, n_features)
            Matrix of contextual features.
        """

        context = self._check_context_matrix(context)
        if context_memory is not None:
            context_memory = self._check_context_matrix(context_memory)
            if context.shape[1] != context_memory.shape[1]:
                raise ValueError("Context memory must have the same number of features as the context.")
        super().update(
            actions=actions,
            rewards=rewards,
            quantities=quantities,
            context=context,
            actions_memory=actions_memory,
            rewards_memory=rewards_memory,
            context_memory=context_memory,
        )

    @staticmethod
    @validate_call(config=dict(arbitrary_types_allowed=True))
    def _check_context_matrix(context: ArrayLike):
        """
        Check and cast context matrix.

        Parameters
        ----------
        context : np.ndarray of shape (n_samples, n_features)
            Matrix of contextual features.

        Returns
        -------
        context : pandas DataFrame of shape (n_samples, n_features)
            Matrix of contextual features.
        """
        try:
            context = np.asarray(context, dtype=float)
        except Exception as e:
            raise AttributeError(f"Context must be an ArrayLike that can transform to float numpy array: {e}.")
        return context

    def _update_actions(
        self,
        actions: List[ActionId],
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]],
        quantities: Optional[List[Union[float, List[float], None]]],
        context: np.ndarray,
    ):
        """
        Update the models associated with the given actions using the provided rewards.

        Parameters
        ----------
        actions : List[UnifiedActionId] of shape (n_samples,), e.g. ['a1', 'a2', 'a3', 'a4', 'a5']
            The selected action for each sample.
        rewards : List[Union[BinaryReward, List[BinaryReward]]] of shape (n_samples, n_objectives)
            The binary reward for each sample.
                If strategy is not MultiObjectiveBandit, rewards should be a list, e.g.
                    rewards = [1, 0, 1, 1, 1, ...]
                If strategy is MultiObjectiveBandit, rewards should be a list of list, e.g. (with n_objectives=2):
                    rewards = [[1, 1], [1, 0], [1, 1], [1, 0], [1, 1], ...]
        quantities : Optional[List[Union[float, List[float], None]]]
            The value associated with each action. If none, the value is not used, i.e. non-quantitative action.
        context: np.ndarray of shape (n_samples, n_features)
            Matrix of contextual features.
        """
        context = context[-len(actions) :]

        rewards_dict = defaultdict(list)
        context_dict = defaultdict(list)
        if quantities is None:
            for a, r, c in zip(actions, rewards, context):
                rewards_dict[a].append(r)
                context_dict[a].append(c)

            for a in set(actions):
                self.actions[a].update(rewards=rewards_dict[a], context=np.array(context_dict[a]))

        else:
            quantities = quantities[-len(actions) :]
            quantities_dict = defaultdict(list)
            for a, r, c, q in zip(actions, rewards, context, quantities):
                rewards_dict[a].append(r)
                context_dict[a].append(c)
                quantities_dict[a].append(q)

            for a in set(actions):
                if any(quantities_dict[a]):  # quantitative action
                    self.actions[a].update(
                        context=np.array(context_dict[a]), rewards=rewards_dict[a], quantities=quantities_dict[a]
                    )
                else:  # non-quantitative action
                    self.actions[a].update(context=np.array(context_dict[a]), rewards=rewards_dict[a])


# For pickling purposes
SmabActionsManagerSO = SmabActionsManager[Union[Beta, SmabZoomingModel]]
SmabActionsManagerCC = SmabActionsManager[Union[BetaCC, SmabZoomingModelCC]]
SmabActionsManagerMO = SmabActionsManager[BetaMO]
SmabActionsManagerMOCC = SmabActionsManager[BetaMOCC]

CmabActionsManagerSO = CmabActionsManager[Union[BayesianNeuralNetwork, CmabZoomingModel]]
CmabActionsManagerCC = CmabActionsManager[Union[BayesianNeuralNetworkCC, CmabZoomingModelCC]]



# MIT License
#
# Copyright (c) 2022 Playtika Ltd.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import inspect
from abc import ABC
from types import ModuleType
from typing import Callable, List, Optional, Tuple

from bokeh.io import curdoc, output_file, output_notebook, save, show
from bokeh.models import InlineStyleSheet, TabPanel, Tabs
from IPython import get_ipython

from pybandits.pydantic_version_compatibility import validate_call


@validate_call
def extract_argument_names_from_function(handle: Callable, ignore_arguments: Tuple = ("self", "cls")) -> List[str]:
    """
    Extract the argument names from a function handle.

    Parameters
    ----------
    handle : Callable
        Handle of a function or class to extract the argument names from
    ignore_arguments : Tuple
        Tuple of argument names to ignore

    Returns
    -------
    argument_names : List[str]
        List of argument names
    """

    argument_names = list(
        handle.model_fields.keys() if hasattr(handle, "model_fields") else inspect.signature(handle).parameters
    )
    for argument_name in ignore_arguments:
        if argument_name in argument_names:
            argument_names.remove(argument_name)
    return argument_names


@validate_call(config=dict(arbitrary_types_allowed=True))
def get_non_abstract_classes(module: ModuleType) -> List[type]:
    non_abc_classes = []
    for name, obj in inspect.getmembers(module, inspect.isclass):
        # Ensure the class is defined in the module and not imported
        if obj.__module__ == module.__name__:
            # Check if the class is not an abstract class (i.e., doesn't inherit from abc.ABC)
            if not inspect.isabstract(obj) and ABC not in obj.__bases__:
                non_abc_classes.append(obj)
    return non_abc_classes


def in_jupyter_notebook() -> bool:
    """
    Check if the code is running in a Jupyter notebook.

    References
    ----------
    https://stackoverflow.com/a/39662359

    Returns
    -------
    bool
        True if the code is running in a Jupyter notebook, False otherwise.

    Raises
    ------
    NotImplementedError
        If the shell type is neither Jupyter notebook nor terminal.
    """

    try:
        shell = get_ipython().__class__.__name__

        return shell == "ZMQInteractiveShell"

    except NameError:
        return False  # Probably standard Python interpreter


def visualize_via_bokeh(output_path: Optional[str], tabs: List[TabPanel]):
    """
    Visualize output to either a Jupyter notebook or an HTML file.

    Parameters
    ----------
    output_path : Optional[str]
        Path to the output file. Required if not running in a Jupyter notebook.
    tabs : List[TabPanel]
        List of TabPanel objects to visualize.
    """

    if in_jupyter_notebook():
        output_notebook()
    else:
        if output_path is None:
            raise ValueError("output_path is required when not running in a Jupyter notebook.")
        output_file(output_path)

    # Add a Div model to the Bokeh layout for flexible tabs
    tabs_css = """
                 :host(.bk-Tabs) .bk-header {
                     flex-wrap: wrap !important;
                 }
             """

    tabs_stylesheet = InlineStyleSheet(css=tabs_css)
    curdoc().title = "Visual report"
    styled_tabs = Tabs(tabs=tabs, stylesheets=[tabs_stylesheet], sizing_mode="stretch_both")
    if in_jupyter_notebook():
        show(styled_tabs)
    else:
        save(styled_tabs)


class classproperty(property):
    def __get__(self, instance, owner):
        return self.fget(owner)



# MIT License
#
# Copyright (c) 2023 Playtika Ltd.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from abc import ABC, abstractmethod
from copy import deepcopy
from random import betavariate
from typing import ClassVar, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import pytensor.tensor as pt
from numpy import sqrt
from numpy.typing import ArrayLike
from pymc import Bernoulli, Data, Deterministic, fit, math, sample, sample_prior_predictive
from pymc import Model as PymcModel
from pymc import StudentT as PymcStudentT
from pytensor.tensor import specify_broadcastable
from typing_extensions import Self

from pybandits.base import BinaryReward, MOProbability, Probability, ProbabilityWeight, PyBanditsBaseModel
from pybandits.base_model import BaseModelCC, BaseModelMO, BaseModelSO
from pybandits.pydantic_version_compatibility import (
    PYDANTIC_VERSION_1,
    PYDANTIC_VERSION_2,
    Field,
    NonNegativeFloat,
    PositiveFloat,
    PositiveInt,
    PrivateAttr,
    field_validator,
    model_validator,
    pydantic_version,
    validate_call,
)

UpdateMethods = Literal["VI", "MCMC"]


class Model(BaseModelSO, ABC):
    """
    Class to model the prior distributions for single objective.

    Parameters
    ----------
    n_successes: PositiveInt = 1
        Counter of the number of successes.
    n_failures: PositiveInt = 1
        Counter of the number of failures.
    """

    @abstractmethod
    def sample_proba(self, **kwargs) -> Union[List[Probability], List[MOProbability], List[ProbabilityWeight]]:
        """
        Sample the probability of getting a positive reward.
        """


class ModelCC(BaseModelCC, ABC):
    """
    Class to model action cost.

    Parameters
    ----------
    cost: NonNegativeFloat
        Cost associated to the action.
    """

    cost: NonNegativeFloat


class ModelMO(BaseModelMO, ABC):
    """
    Class to model the prior distributions for multi-objective.

    Parameters
    ----------
    models : List[Model]
        The list of models for each objective.
    """

    models: List[Model]


class BaseBeta(Model, ABC):
    """
    Beta Distribution model for Bernoulli multi-armed bandits.

    Parameters
    ----------
    n_successes: PositiveInt = 1
        Counter of the number of successes.
    n_failures: PositiveInt = 1
        Counter of the number of failures.
    """

    @property
    def std(self) -> float:
        """
        The corrected standard deviation (Bessel's correction) of the binary distribution of successes and failures.
        """
        return sqrt((self.n_successes * self.n_failures) / (self.count * (self.count - 1)))

    @validate_call
    def _update(self, rewards: List[BinaryReward]):
        """
        Update n_successes and n_failures.

        Parameters
        ----------
        rewards: List[BinaryReward]
            A list of binary rewards.
        """
        pass

    def _reset(self):
        pass

    def sample_proba(self, n_samples: PositiveInt) -> List[Probability]:
        """
        Sample the probability of getting a positive reward.

        Returns
        -------
        prob: Probability
            Probability of getting a positive reward.
        """
        return [betavariate(self.n_successes, self.n_failures) for _ in range(n_samples)]


class Beta(BaseBeta):
    """
    Beta Distribution model for Bernoulli multi-armed bandits.

    Parameters
    ----------
    n_successes: PositiveInt = 1
        Counter of the number of successes.
    n_failures: PositiveInt = 1
        Counter of the number of failures.
    """


class BetaCC(BaseBeta, ModelCC):
    """
    Beta Distribution model for Bernoulli multi-armed bandits with cost control.

    Parameters
    ----------
    n_successes : PositiveInt = 1
        Counter of the number of successes.
    n_failures : PositiveInt = 1
        Counter of the number of failures.
    cost : NonNegativeFloat
        Cost associated to the Beta distribution.
    """


class BaseBetaMO(ModelMO, ABC):
    """
    Base beta Distribution model for Bernoulli multi-armed bandits with multi-objectives.

    Parameters
    ----------
    models: List[Beta] of length (n_objectives,)
        List of Beta distributions.
    """

    models: List[Beta]

    @validate_call
    def sample_proba(self, n_samples: PositiveInt) -> List[MOProbability]:
        """
        Sample the probability of getting a positive reward.

        Parameters
        ----------
        n_samples : PositiveInt
            Number of samples to draw.

        Returns
        -------
        prob: List[MOProbability]
            Probabilities of getting a positive reward for each sample and objective.
        """
        return [list(p) for p in zip(*[model.sample_proba(n_samples=n_samples) for model in self.models])]

    @validate_call
    def _update(self, rewards: List[List[BinaryReward]]):
        """
        Update the Beta model using the provided rewards.

        Parameters
        ----------
        rewards: List[List[BinaryReward]]
            A list of rewards, where each reward is in turn a list containing the reward of the Beta model
            associated to each objective.
            For example, `[[1, 1], [1, 0], [1, 1], [1, 0], [1, 1]]`.
        """
        if any(len(x) != len(self.models) for x in rewards):
            raise AttributeError("The shape of rewards is incorrect")

        for i, model in enumerate(self.models):
            model.update([r[i] for r in rewards])

    @classmethod
    def cold_start(cls, n_objectives: PositiveInt, **kwargs) -> "BetaMO":
        """
        Utility function to create a Bayesian Logistic Regression model  or child model with cost control,
        with default parameters.

        It is modeled as:

            y = sigmoid(alpha + beta1 * x1 + beta2 * x2 + ... + betaN * xN)

        where the alpha and betas coefficients are Student's t-distributions.

        Parameters
        ----------
        n_betas : PositiveInt
            The number of betas of the Bayesian Logistic Regression model. This is also the number of features expected
            after in the context matrix.
        kwargs: Dict[str, Any]
            Additional arguments for the Bayesian Logistic Regression child model.

        Returns
        -------
        beta_mo: BetaMO
            The multi-objective Beta model.
        """
        models = n_objectives * [Beta()]
        beta_mo = cls(models=models, **kwargs)
        return beta_mo


class BetaMO(BaseBetaMO):
    """
    Beta Distribution model for Bernoulli multi-armed bandits with multi-objectives.

    Parameters
    ----------
    models: List[Beta] of length (n_objectives,)
        List of Beta distributions.
    """


class BetaMOCC(BaseBetaMO, ModelCC):
    """
    Beta Distribution model for Bernoulli multi-armed bandits with multi-objectives and cost control.

    Parameters
    ----------
    models: List[BetaCC] of shape (n_objectives,)
        List of Beta distributions.
    cost: NonNegativeFloat
        Cost associated to the Beta distribution.
    """


class StudentTArray(PyBanditsBaseModel):
    """
    A class representing an array of Student's t-distributions with parameters `mu`, `sigma`, and `nu`.
    A specific element (e.g, a single parameter of a layer) distribution is defined by the the corresponding elements in the lists.
    The mean values are represented by `mu`, the scale (standard deviation) values by `sigma`, and the degrees of freedom by `nu`.


    Parameters
    ----------
    mu : Union[List[float], List[List[float]]]
        The mean values of the Student's t-distributions. Can be a 1D (for the layer bias term) or 2D list (for the layer weight term).
    sigma : Union[List[NonNegativeFloat], List[List[NonNegativeFloat]]]
        The scale (standard deviation) values of the Student's t-distributions. Must be non-negative.
        Can be a 1D or 2D list.
    nu : Union[List[PositiveFloat], List[List[PositiveFloat]]]
        The degrees of freedom of the Student's t-distributions. Must be positive.
        Can be a 1D or 2D list.
    """

    mu: Union[List[float], List[List[float]]]
    sigma: Union[List[NonNegativeFloat], List[List[NonNegativeFloat]]]
    nu: Union[List[PositiveFloat], List[List[PositiveFloat]]]

    @staticmethod
    def convert_list_to_array(input_list: Union[List[float], List[List[float]]]) -> bool:
        if len(input_list) == 0:
            is_valid_input = False

        elif not isinstance(input_list[0], list):
            is_valid_input = True

        else:
            first_length = len(input_list[0])
            is_valid_input = all(
                isinstance(inner_list, list) and len(inner_list) == first_length for inner_list in input_list
            )

        if is_valid_input:
            return np.array(input_list)
        else:
            raise ValueError("Input list must be a 1D or 2D list with the same length for all inner lists.")

    @model_validator(mode="after")
    @classmethod
    def validate_inputs(cls, values):
        if pydantic_version == PYDANTIC_VERSION_1:
            mu_arr = cls.convert_list_to_array(values.get("mu"))
            sigma_arr = cls.convert_list_to_array(values.get("sigma"))
            nu_arr = cls.convert_list_to_array(values.get("nu"))
        elif pydantic_version == PYDANTIC_VERSION_2:
            mu_arr = cls.convert_list_to_array(values.mu)
            sigma_arr = cls.convert_list_to_array(values.sigma)
            nu_arr = cls.convert_list_to_array(values.nu)
        else:
            raise ValueError(f"Unsupported pydantic version: {pydantic_version}")

        if (mu_arr.shape != sigma_arr.shape) or (mu_arr.shape != nu_arr.shape):
            raise ValueError(
                f"mu, sigma, and nu must have the same shape, "
                f"but are {mu_arr.shape}, {sigma_arr.shape}, and {nu_arr.shape}, respectively."
            )

        if any(dim_len == 0 for dim_len in mu_arr.shape):
            raise ValueError("mu, sigma, and nu must have at least one element in every dimension.")

        return values

    @classmethod
    def cold_start(
        cls,
        shape: Union[PositiveInt, Tuple[PositiveInt, ...]],
        mu: float = 0.0,
        sigma: NonNegativeFloat = 10.0,
        nu: PositiveFloat = 5.0,
    ) -> "StudentTArray":
        if isinstance(shape, int):
            shape = (shape,)

        if any(dim_len == 0 for dim_len in shape):
            raise ValueError("shape of mu, sigma, and nu must have at least one element in every dimension.")

        mu = np.full(shape, mu).tolist()
        sigma = np.full(shape, sigma).tolist()
        nu = np.full(shape, nu).tolist()
        return cls(mu=mu, sigma=sigma, nu=nu)

    @property
    def shape(self) -> Tuple[PositiveInt, ...]:
        return np.array(self.mu).shape

    @property
    def params(self):
        return dict(mu=np.array(self.mu), sigma=np.array(self.sigma), nu=np.array(self.nu))


class BnnLayerParams(PyBanditsBaseModel):
    """
    Represents the parameters of a Bayesian neural network (BNN) layer.

    Parameters
    ----------
    weight : StudentTArray
        The weight parameter of the BNN layer, represented as a StudentTArray.
    bias : StudentTArray
        The bias parameter of the BNN layer, represented as a StudentTArray.
    """

    weight: StudentTArray
    bias: StudentTArray


class BnnParams(PyBanditsBaseModel):
    """
    Represents the parameters of a Bayesian Neural Network (BNN), including
    both the current layer parameters and the initial layer parameters.
    We keep the init parameters in case we need to reset the model.

    Parameters
    ----------
    bnn_layer_params : List[BnnLayerParams]
        A list of BNN layer parameters representing the current state of the model.
    bnn_layer_params_init : List[BnnLayerParams]
        A list of BNN layer parameters representing the initial state of the model.
    """

    bnn_layer_params: Optional[List[BnnLayerParams]]
    bnn_layer_params_init: List[BnnLayerParams] = Field(default_factory=list, init=False, frozen=True)

    @model_validator(mode="before")
    @classmethod
    def validate_inputs(cls, values):
        if values.get("bnn_layer_params_init") is None:
            values["bnn_layer_params_init"] = deepcopy(values["bnn_layer_params"])

        return values


class BaseBayesianNeuralNetwork(Model, ABC):
    """Bayesian Neural Network model for binary classification.

    This class implements a Bayesian Neural Network with an arbitrary number of fully connected layers
    using PyMC for binary classification tasks. It supports both Markov Chain Monte Carlo (MCMC)
    and Variational Inference (VI) methods for posterior inference.

    References
    ----------
    Bayesian Learning for Neural Networks (Radford M. Neal, 1995)
    https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=db869fa192a3222ae4f2d766674a378e47013b1b

    Parameters
    ----------
    model_params : BnnParams
        The parameters of the Bayesian Neural Network, including weights and biases for each layer and their initial values for resetting
    update_method : str, optional
        The method used for posterior inference, either "MCMC" or "VI" (default is "MCMC").
    update_kwargs : Optional[dict], optional
        A dictionary of keyword arguments for the update method. For MCMC, it contains 'trace' settings.
        For VI, it contains both 'trace' and 'fit' settings.

    Notes
    -----
    - The model uses tanh activation for hidden layers and sigmoid activation for the output layer.
    - The output layer is designed for binary classification tasks, with probabilities modeled
      using a Bernoulli likelihood.
    """

    model_params: BnnParams

    _logit_var_name: ClassVar[str] = "logit"
    _prob_var_name: ClassVar[str] = "prob"
    _weight_var_name: ClassVar[str] = "weight"
    _bias_var_name: ClassVar[str] = "bias"

    update_method: str = "MCMC"
    update_kwargs: Optional[dict] = None

    _default_mcmc_trace_kwargs: ClassVar[dict] = dict(
        tune=500,
        draws=1000,
        chains=2,
        init="adapt_diag",
        cores=1,
        target_accept=0.95,
        progressbar=False,
        return_inferencedata=False,
    )

    _default_variational_inference_fit_kwargs: ClassVar[dict] = dict(method="advi")
    _default_variational_inference_trace_kwargs: ClassVar[dict] = dict(
        draws=1000, progressbar=False, return_inferencedata=False
    )

    _approx_history: np.ndarray = PrivateAttr(None)

    class Config:
        arbitrary_types_allowed = True

    if pydantic_version == PYDANTIC_VERSION_1:

        @model_validator(mode="before")
        @classmethod
        def arrange_update_kwargs(cls, values):
            update_kwargs = cls._get_value_with_default("update_kwargs", values)
            update_method = cls._get_value_with_default("update_method", values)

            if update_kwargs is None:
                update_kwargs = dict()

            if update_method == "VI":
                update_kwargs["trace"] = {
                    **cls._default_variational_inference_trace_kwargs,
                    **update_kwargs.get("trace", {}),
                }
                update_kwargs["fit"] = {**cls._default_variational_inference_fit_kwargs, **update_kwargs.get("fit", {})}

            elif update_method == "MCMC":
                update_kwargs["trace"] = {**cls._default_mcmc_trace_kwargs, **update_kwargs.get("trace", {})}
            else:
                raise ValueError("Invalid update method.")

            values["update_kwargs"] = update_kwargs
            values["update_method"] = update_method
            return values

    elif pydantic_version == PYDANTIC_VERSION_2:

        @model_validator(mode="after")
        def arrange_update_kwargs(self):
            if self.update_kwargs is None:
                self.update_kwargs = dict()

            if self.update_method == "VI":
                self.update_kwargs["trace"] = {
                    **self._default_variational_inference_trace_kwargs,
                    **self.update_kwargs.get("trace", {}),
                }
                self.update_kwargs["fit"] = {
                    **self._default_variational_inference_fit_kwargs,
                    **self.update_kwargs.get("fit", {}),
                }

            elif self.update_method == "MCMC":
                self.update_kwargs["trace"] = {**self._default_mcmc_trace_kwargs, **self.update_kwargs.get("trace", {})}
            else:
                raise ValueError("Invalid update method.")
            return self

    else:
        raise ValueError(f"Unsupported pydantic version: {pydantic_version}")

    @property
    def approx_history(self) -> Optional[np.ndarray]:
        return self._approx_history

    @classmethod
    def get_layer_params_name(cls, layer_ind: PositiveInt) -> Tuple[str, str]:
        weight_layer_params_name = f"{cls._weight_var_name}_{layer_ind}"
        bias_layer_params_name = f"{cls._bias_var_name}_{layer_ind}"
        return weight_layer_params_name, bias_layer_params_name

    @classmethod
    def create_model_params(
        cls, n_features: PositiveInt, hidden_dim_list: List[PositiveInt], **dist_params_init
    ) -> BnnParams:
        """
        Creates model parameters for a Bayesian neural network (BNN) model according to dist_params_init
        This method initializes the distribution's parameters for each layer of a BNN
        using the specified number of features, hidden dimensions, and distribution
        initialization parameters.

        Parameters
        ----------
        n_features : PositiveInt
            The number of input features for the BNN.
        hidden_dim_list : List[PositiveInt]
            A list of integers specifying the number of hidden units in each hidden layer.
            If None, no hidden layers are added.
        **dist_params_init : dict, optional
            Additional parameters for initializing the distribution of weights and biases.
        Returns
        -------
        BnnParams
            An instance of BnnParams containing the initialized layer parameters.
        """

        if hidden_dim_list is None:
            _dim_list = [n_features]
        else:
            _dim_list = [n_features] + hidden_dim_list

        _dim_list.append(1)

        layer_params_init = []
        for layer_ind in range(len(_dim_list) - 1):
            input_dim = _dim_list[layer_ind]
            output_dim = _dim_list[layer_ind + 1]
            w_param = StudentTArray.cold_start(shape=(input_dim, output_dim), **dist_params_init)
            b_param = StudentTArray.cold_start(shape=output_dim, **dist_params_init)
            layer_params_init.append(BnnLayerParams(weight=w_param, bias=b_param))

        return BnnParams(bnn_layer_params=layer_params_init)

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def check_context_matrix(self, context: np.ndarray):
        """
        Check and cast context matrix.

        Parameters
        ----------
        context : np.ndarray of shape (n_samples, n_features)
            Matrix of contextual features.

        Returns
        -------
        context : pandas DataFrame of shape (n_samples, n_features)
            Matrix of contextual features.
        """
        try:
            n_cols_context = np.array(context).shape[1]
        except Exception as e:
            raise AttributeError(f"Context must be an ArrayLike with {self.input_dim} columns: {e}.")

        if not np.issubdtype(context.dtype, np.number):
            raise ValueError("Context array must contain only numeric values.")

        if n_cols_context != self.input_dim:
            raise AttributeError(f"Shape mismatch: context must have {self.input_dim} columns.")

    @property
    def input_dim(self) -> PositiveInt:
        """
        Returns the expected input dimension of the model.

        Returns
        -------
        int
            The number of input features expected by the model, derived from
            the shape of the weight matrix in the first layer's parameters.
        """
        return self.model_params.bnn_layer_params[0].weight.shape[0]

    def create_model(
        self, x: ArrayLike, y: Optional[Union[List[BinaryReward], np.ndarray]] = None, is_predict: bool = False
    ) -> PymcModel:
        """
        Create a PyMC model for Bayesian Neural Network.

        This method builds a PyMC model with the network architecture specified in model_params.
        The model uses tanh activation for hidden layers and sigmoid for the output layer.

        Parameters
        ----------
        x : ArrayLike
            Input features of shape (n_samples, n_features)
        y : Union[List[BinaryReward], np.ndarray]
            Binary target values of shape (n_samples,)
        is_predict : bool
            If True, process samples independently. If False, process all samples at once.
            In the predict step, we would like to sample the model parameters independently for each sample.
            In the update step, this is not required.

        Returns
        -------
        PymcModel
            PyMC model object with the specified neural network architecture

        Notes
        -----
        The model structure follows these steps:
        1. For each layer, create weight and bias variables from StudentT distributions.
        2. Apply linear transformations and activations through the layers.
           When is_sampelwise is True, the linear transformation is applied on each row separately (so random variables are not shared).
           When is_sampelwise is False, the linear transformation is applied on the whole matrix at once, so random variables are shared.
        3. Apply sigmoid activation at the output
        4. Use Bernoulli likelihood for binary classification
        """
        if y is None:
            y = np.zeros(len(x), dtype=np.int64)

        with PymcModel() as _model:
            # Define data variables using minibatches
            bnn_output = Data("bnn_output", y, mutable=True)
            if is_predict:
                # arrange  for batched dot product
                x_expanded = np.expand_dims(x, axis=1)
                bnn_input = Data("bnn_input", x_expanded, mutable=True)
                for dim, size in enumerate(x_expanded.shape):
                    if size == 1:
                        bnn_input = specify_broadcastable(bnn_input, dim)
            else:
                bnn_input = Data("bnn_input", x, mutable=True)

            n_samples = len(x)

            next_layer_input = bnn_input
            for layer_ind, layer_params in enumerate(self.model_params.bnn_layer_params):
                w_shape = layer_params.weight.shape  # without it n_features = 1 doesn't work
                b_shape = layer_params.bias.shape
                weight_layer_params_name, bias_layer_params_name = self.get_layer_params_name(layer_ind)

                if is_predict:
                    # in this case we create n_samples different weights and biases - one for each sample
                    w = PymcStudentT(
                        name=weight_layer_params_name, shape=(n_samples,) + w_shape, **layer_params.weight.params
                    )
                    b = PymcStudentT(
                        name=bias_layer_params_name, shape=(n_samples, 1) + b_shape, **layer_params.bias.params
                    )

                    batched_dot = pt.vectorize(
                        lambda x_batch, w_batch: pt.dot(x_batch, w_batch), signature="(n,m),(m,p)->(n,p)"
                    )
                    linear_transform = pt.as_tensor_variable(
                        batched_dot(next_layer_input, w) + b, name=f"linear_transform{layer_ind}"
                    )

                else:
                    # in this case we create one weight and bias for all samples
                    w = PymcStudentT(
                        name=weight_layer_params_name, shape=w_shape, **layer_params.weight.params, initval="prior"
                    )
                    b = PymcStudentT(
                        name=bias_layer_params_name, shape=b_shape, **layer_params.bias.params, initval="prior"
                    )

                    linear_transform = Deterministic(f"linear_transform{layer_ind}", math.dot(next_layer_input, w) + b)

                if layer_ind < len(self.model_params.bnn_layer_params) - 1:
                    next_layer_input = math.tanh(linear_transform)
            for dim in range(1, linear_transform.ndim):
                linear_transform = specify_broadcastable(linear_transform, dim)

            logit = Deterministic(self._logit_var_name, linear_transform.squeeze())
            Deterministic(self._prob_var_name, math.sigmoid(logit))

            Bernoulli("out", logit_p=logit, observed=bnn_output)
        return _model

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def sample_proba(self, context: np.ndarray) -> List[ProbabilityWeight]:
        """
        Samples probabilities and weighted sums from the prior predictive distribution.

        Parameters
        ----------
        context : ArrayLike
            The context matrix for which the probabilities are to be sampled.
        Returns
        -------
        List[ProbabilityWeight]
            Each element is a tuple containing the probability of a positive reward and
            the corresponding weighted sum between contextual feature quantities and sampled coefficients.
        """

        # check input args
        self.check_context_matrix(context=context)

        _context = np.atleast_2d(context)
        _model = self.create_model(x=_context, is_predict=True)

        with _model:
            trace = sample_prior_predictive(samples=1)

        prob = trace["prior"][self._prob_var_name].values.reshape(-1)
        weighted_sum = trace["prior"][self._logit_var_name].values.reshape(-1)

        return list(zip(prob, weighted_sum))

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def _update(self, context: np.ndarray, rewards: List[BinaryReward]):
        """
        Update the model_params with new context and rewards.
        Parameters
        ----------
        context : np.ndarray
            The context matrix where each row represents a context vector.
        rewards : List[BinaryReward]
            A list of binary rewards corresponding to each context vector.

        Notes
        -----
        This method updates the model's parameters by sampling from the posterior distribution
        using either Variational Inference (VI) or Markov Chain Monte Carlo (MCMC) methods.
        """
        self.check_context_matrix(context=context)

        if len(context) != len(rewards):
            raise AttributeError("Shape mismatch: context and rewards must have the same length.")

        _context = np.atleast_2d(context)
        _model = self.create_model(x=_context, y=rewards, is_predict=False)
        with _model:
            # update traces object by sampling from posterior distribution
            if self.update_method == "VI":
                # variational inference
                update_kwargs = self.update_kwargs.copy()
                approx = fit(**update_kwargs["fit"])
                trace = approx.sample(**update_kwargs["trace"])
                self._approx_history = approx.hist
            elif self.update_method == "MCMC":
                # MCMC
                trace = sample(**self.update_kwargs["trace"])
            else:
                raise ValueError("Invalid update method.")

        for layer_ind, layer_params in enumerate(self.model_params.bnn_layer_params):
            weight_layer_params_name, bias_layer_params_name = self.get_layer_params_name(layer_ind)

            w_mu = np.mean(trace[weight_layer_params_name], axis=0)
            w_sigma = np.std(trace[weight_layer_params_name], axis=0)
            layer_params.weight = StudentTArray(
                mu=w_mu.tolist(), sigma=w_sigma.tolist(), nu=self.model_params.bnn_layer_params[layer_ind].weight.nu
            )

            b_mu = np.mean(trace[bias_layer_params_name], axis=0)
            b_sigma = np.std(trace[bias_layer_params_name], axis=0)
            layer_params.bias = StudentTArray(
                mu=b_mu.tolist(), sigma=b_sigma.tolist(), nu=self.model_params.bnn_layer_params[layer_ind].bias.nu
            )

    @classmethod
    def cold_start(
        cls,
        n_features: PositiveInt,
        hidden_dim_list: Optional[List[PositiveInt]] = None,
        update_method: UpdateMethods = "MCMC",
        update_kwargs: Optional[dict] = None,
        dist_params_init: Optional[Dict[str, float]] = None,
        **kwargs,
    ) -> Self:
        """
        Initialize a Bayesian Neural Network with a cold start.

        Parameters
        ----------
        n_features : PositiveInt
            Number of input features for the network.
        hidden_dim_list : Optional[List[PositiveInt]], optional
            List of dimensions for the hidden layers of the network. If None, no hidden layers are added.
        update_method : UpdateMethods
            Method to update the network, either "MCMC" or "VI". Default is "MCMC".
        update_kwargs : Optional[dict], optional
            Additional keyword arguments for the update method. Default is None.
        dist_params_init : Optional[Dict[str, float]], optional
            Initial distribution parameters for the network weights and biases. Default is None.
        **kwargs
            Additional keyword arguments for the BayesianNeuralNetwork constructor.

        Returns
        -------
        Self
            An instance of the Bayesian Neural Network initialized with the specified parameters.
        """

        if dist_params_init is None:
            dist_params_init = {}

        model_params = cls.create_model_params(
            n_features=n_features, hidden_dim_list=hidden_dim_list, **dist_params_init
        )
        return cls(model_params=model_params, update_method=update_method, update_kwargs=update_kwargs, **kwargs)

    def _reset(self):
        """
        Reset the model.
        """
        self.model_params.bnn_layer_params = deepcopy(self.model_params.bnn_layer_params_init)


class BayesianNeuralNetwork(BaseBayesianNeuralNetwork):
    """
    Bayesian Neural Network class.
    This class implements a Bayesian Neural Network by extending the
    BaseBayesianNeuralNetwork. It provides functionality for probabilistic
    modeling and inference using neural networks.
    """


class BayesianNeuralNetworkCC(BaseBayesianNeuralNetwork, ModelCC):
    """Bayesian Neural Network model for binary classification with cost constraint.

    This class implements a Bayesian Neural Network with an arbitrary number of fully connected layers
    using PyMC for binary classification tasks. It supports both Markov Chain Monte Carlo (MCMC)
    and Variational Inference (VI) methods for posterior inference.

    References
    ----------
    Bayesian Learning for Neural Networks (Radford M. Neal, 1995)
    https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=db869fa192a3222ae4f2d766674a378e47013b1b

    Parameters
    ----------
    model_params : BnnParams
        The parameters of the Bayesian Neural Network, including weights and biases for each layer and their initial values for resetting
    update_method : str, optional
        The method used for posterior inference, either "MCMC" or "VI" (default is "MCMC").
    update_kwargs : Optional[dict], optional
        A dictionary of keyword arguments for the update method. For MCMC, it contains 'trace' settings.
        For VI, it contains both 'trace' and 'fit' settings.
    cost : NonNegativeFloat
        Cost associated to the Bayesian Neural Network model.

    Notes
    -----
    - The model uses tanh activation for hidden layers and sigmoid activation for the output layer.
    - The output layer is designed for binary classification tasks, with probabilities modeled
      using a Bernoulli likelihood.
    """


class BayesianLogisticRegression(BayesianNeuralNetwork):
    """
    A Bayesian Logistic Regression model that inherits from BayesianNeuralNetwork.
    This model is a specialized version of a Bayesian Neural Network with a single layer,
    designed specifically for logistic regression tasks. The model parameters are
    validated to ensure that the model adheres to this single-layer constraint.
    """

    @field_validator("model_params")
    def validate_model_params(cls, model_params):
        if (len(model_params.bnn_layer_params_init) != 1) or (len(model_params.bnn_layer_params) != 1):
            raise ValueError("The Bayesian Logistic Regression model should have only one layer.")
        return model_params


class BayesianLogisticRegressionCC(BayesianLogisticRegression, ModelCC):
    """
    A Bayesian Logistic Regression model with cost control.
    """



# MIT License
#
# Copyright (c) 2022 Playtika Ltd.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import ABC
from typing import List, Optional, Set, Union

from pybandits.actions_manager import (
    SmabActionsManager,
    SmabActionsManagerCC,
    SmabActionsManagerMO,
    SmabActionsManagerMOCC,
    SmabActionsManagerSO,
)
from pybandits.base import (
    ActionId,
    BinaryReward,
    SmabPredictions,
)
from pybandits.mab import BaseMab
from pybandits.model import BaseBeta
from pybandits.pydantic_version_compatibility import PositiveInt, validate_call
from pybandits.quantitative_model import BaseSmabZoomingModel
from pybandits.strategy import (
    BestActionIdentificationBandit,
    ClassicBandit,
    CostControlBandit,
    MultiObjectiveBandit,
    MultiObjectiveCostControlBandit,
)


class BaseSmabBernoulli(BaseMab, ABC):
    """
    Base model for a Stochastic Bernoulli Multi-Armed Bandit with Thompson Sampling.

    Parameters
    ----------
    actions: Dict[ActionId, Union[BaseBeta, BaseSmabZoomingModel]]
        The list of possible actions, and their associated Model.
    strategy: Strategy
        The strategy used to select actions.
    """

    actions_manager: SmabActionsManager[Union[BaseBeta, BaseSmabZoomingModel]]

    @validate_call
    def predict(
        self,
        n_samples: PositiveInt = 1,
        forbidden_actions: Optional[Set[ActionId]] = None,
    ) -> SmabPredictions:
        """
        Predict actions.

        Parameters
        ----------
        n_samples : PositiveInt, default=1
            Number of samples to predict.
        forbidden_actions : Optional[Set[ActionId]], default=None
            Set of forbidden actions. If specified, the model will discard the forbidden_actions and it will only
            consider the remaining allowed_actions. By default, the model considers all actions as allowed_actions.
            Note that: actions = allowed_actions U forbidden_actions.

        Returns
        -------
        actions: List[UnifiedActionId]
            The actions selected by the multi-armed bandit model.
        probs: Union[List[Dict[UnifiedActionId, Probability]], List[Dict[UnifiedActionId, MOProbability]]]
            The probabilities of getting a positive reward for each action.
        """

        probs = self._get_action_probabilities(forbidden_actions=forbidden_actions, n_samples=n_samples)
        selected_actions = [self._select_epsilon_greedy_action(p=prob, actions=self.actions) for prob in probs]

        return selected_actions, probs

    @validate_call
    def update(
        self,
        actions: List[ActionId],
        rewards: Union[List[BinaryReward], List[List[BinaryReward]]],
        quantities: Optional[List[Union[float, List[float], None]]] = None,
        actions_memory: Optional[List[ActionId]] = None,
        rewards_memory: Optional[Union[List[BinaryReward], List[List[BinaryReward]]]] = None,
    ):
        """
        Update the stochastic Bernoulli bandit given the list of selected actions and their corresponding binary
        rewards.

        Parameters
        ----------
        actions : List[ActionId] of shape (n_samples,), e.g. ['a1', 'a2', 'a3', 'a4', 'a5']
            The selected action for each sample.
        rewards : List[Union[BinaryReward, List[BinaryReward]]] of shape (n_samples, n_objectives)
            The binary reward for each sample.
                If strategy is not MultiObjectiveBandit, rewards should be a list, e.g.
                    rewards = [1, 0, 1, 1, 1, ...]
                If strategy is MultiObjectiveBandit, rewards should be a list of list, e.g. (with n_objectives=2):
                    rewards = [[1, 1], [1, 0], [1, 1], [1, 0], [1, 1], ...]
        quantities : Optional[List[Union[float, List[float], None]]]
            The value associated with each action. If none, the value is not used, i.e. non-quantitative action.
        actions_memory : Optional[List[ActionId]]
            List of previously selected actions.
        rewards_memory : Optional[Union[List[BinaryReward], List[List[BinaryReward]]]]
            List of previously collected rewards.
        """
        super().update(
            actions=actions,
            rewards=rewards,
            quantities=quantities,
            actions_memory=actions_memory,
            rewards_memory=rewards_memory,
        )


class SmabBernoulli(BaseSmabBernoulli):
    """
    Stochastic Bernoulli Multi-Armed Bandit with Thompson Sampling.

    References
    ----------
    Analysis of Thompson Sampling for the Multi-armed Bandit Problem (Agrawal and Goyal, 2012)
    http://proceedings.mlr.press/v23/agrawal12/agrawal12.pdf

    Parameters
    ----------
    actions_manager: SmabActionsManagerSO
        The manager for actions and their associated models.
    strategy: ClassicBandit
        The strategy used to select actions.
    """

    actions_manager: SmabActionsManagerSO
    strategy: ClassicBandit


class SmabBernoulliBAI(BaseSmabBernoulli):
    """
    Stochastic Bernoulli Multi-Armed Bandit with Thompson Sampling, and Best Action Identification strategy.

    References
    ----------
    Analysis of Thompson Sampling for the Multi-armed Bandit Problem (Agrawal and Goyal, 2012)
    http://proceedings.mlr.press/v23/agrawal12/agrawal12.pdf

    Parameters
    ----------
    actions_manager: SmabActionsManagerSO
        The manager for actions and their associated models.
    strategy: BestActionIdentificationBandit
        The strategy used to select actions.
    """

    actions_manager: SmabActionsManagerSO
    strategy: BestActionIdentificationBandit


class SmabBernoulliCC(BaseSmabBernoulli):
    """
    Stochastic Bernoulli Multi-Armed Bandit with Thompson Sampling, and Cost Control strategy.

    The sMAB is extended to include a control of the action cost. Each action is associated with a predefined "cost".
    At prediction time, the model considers the actions whose expected rewards is above a pre-defined lower bound. Among
    these actions, the one with the lowest associated cost is recommended. The expected reward interval for feasible
    actions is defined as [(1-subsidy_factor) * max_p, max_p], where max_p is the highest expected reward sampled value.

    References
    ----------
    Thompson Sampling for Contextual Bandit Problems with Auxiliary Safety Constraints (Daulton et al., 2019)
    https://arxiv.org/abs/1911.00638

    Multi-Armed Bandits with Cost Subsidy (Sinha et al., 2021)
    https://arxiv.org/abs/2011.01488

    Parameters
    ----------
    actions_manager: SmabActionsManagerCC
        The manager for actions and their associated models.
    strategy: CostControlBandit
        The strategy used to select actions.
    """

    actions_manager: SmabActionsManagerCC
    strategy: CostControlBandit


class SmabBernoulliMO(BaseSmabBernoulli):
    """
    Stochastic Bernoulli Multi-Armed Bandit with Thompson Sampling, and Multi-Objectives strategy.

    The reward pertaining to an action is a multidimensional vector instead of a scalar value. In this setting,
    different actions are compared according to Pareto order between their expected reward vectors, and those actions
    whose expected rewards are not inferior to that of any other actions are called Pareto optimal actions, all of which
    constitute the Pareto front.

    References
    ----------
    Thompson Sampling for Multi-Objective Multi-Armed Bandits Problem (Yahyaa and Manderick, 2015)
    https://www.researchgate.net/publication/272823659_Thompson_Sampling_for_Multi-Objective_Multi-Armed_Bandits_Problem

    Parameters
    ----------
    actions_manager: SmabActionsManagerMO
        The manager for actions and their associated models.
    strategy: MultiObjectiveBandit
        The strategy used to select actions.
    """

    actions_manager: SmabActionsManagerMO
    strategy: MultiObjectiveBandit


class SmabBernoulliMOCC(BaseSmabBernoulli):
    """
    Stochastic Bernoulli Multi-Armed Bandit with Thompson Sampling implementation for Multi-Objective (MO) with Cost
    Control (CC) strategy.

    This Bandit allows the reward to be a multidimensional vector and include a control of the action cost. It merges
    the Multi-Objective and Cost Control strategies.

    Parameters
    ----------
    actions_manager: SmabActionsManagerMOCC
        The manager for actions and their associated models.
    strategy: MultiObjectiveCostControlBandit
        The strategy used to select actions.
    """

    actions_manager: SmabActionsManagerMOCC
    strategy: MultiObjectiveCostControlBandit



from abc import ABC, abstractmethod
from typing import Callable, List, Union

import numpy as np

from pybandits.base import (
    BinaryReward,
    Float01,
    MOProbability,
    Probability,
    ProbabilityWeight,
    PyBanditsBaseModel,
    QuantitativeMOProbability,
    QuantitativeProbability,
    QuantitativeProbabilityWeight,
)
from pybandits.pydantic_version_compatibility import (
    NonNegativeFloat,
    NonNegativeInt,
    PositiveInt,
    validate_call,
)


class BaseModel(PyBanditsBaseModel, ABC):
    """
    Class to model the prior distributions of standard actions and quantitative actions.
    """

    @abstractmethod
    def sample_proba(
        self, **kwargs
    ) -> Union[
        List[Probability],
        List[MOProbability],
        List[ProbabilityWeight],
        List[QuantitativeProbability],
        List[QuantitativeMOProbability],
        List[QuantitativeProbabilityWeight],
    ]:
        """
        Sample the probability of getting a positive reward.
        """

    @abstractmethod
    def update(self, rewards: Union[List[BinaryReward], List[List[BinaryReward]]], **kwargs):
        """
        Update the model parameters.

        Parameters
        ----------
        rewards : Union[List[BinaryReward], List[List[BinaryReward]]],
            if nested list, len() should follow shape of (n_samples, n_objectives)
            The binary reward for each sample.
                If strategy is not MultiObjectiveBandit, rewards should be a list, e.g.
                    rewards = [1, 0, 1, 1, 1, ...]
                If strategy is MultiObjectiveBandit, rewards should be a list of list, e.g. (with n_objectives=2):
                    rewards = [[1, 1], [1, 0], [1, 1], [1, 0], [1, 1], ...]
        """

    @abstractmethod
    def reset(self):
        """
        Reset the model.
        """


class BaseModelSO(BaseModel, ABC):
    """
    Class to model the prior distributions of standard actions and quantitative actions for single objective.

    Parameters
    ----------
    n_successes: PositiveInt = 1
        Counter of the number of successes.
    n_failures: PositiveInt = 1
        Counter of the number of failures.
    """

    n_successes: PositiveInt = 1
    n_failures: PositiveInt = 1

    @abstractmethod
    def sample_proba(
        self, **kwargs
    ) -> Union[
        List[Probability], List[ProbabilityWeight], List[QuantitativeProbability], List[QuantitativeProbabilityWeight]
    ]:
        """
        Sample the probability of getting a positive reward.
        """

    @validate_call(config=dict(arbitrary_types_allowed=True))
    def update(self, rewards: List[BinaryReward], **kwargs):
        """
        Update the model parameters.

        Parameters
        ----------
        rewards : List[BinaryReward],
            The binary reward for each sample.
        """
        self._update(rewards=rewards, **kwargs)
        self.n_successes += sum(rewards)
        self.n_failures += len(rewards) - sum(rewards)

    @abstractmethod
    def _update(self, rewards: List[BinaryReward], **kwargs):
        """
        Update the model parameters.

        Parameters
        ----------
        rewards: List[BinaryReward]
            A list of binary rewards.
        """

    def reset(self):
        """
        Reset the model.
        """
        self.n_successes = 1
        self.n_failures = 1
        self._reset()

    @abstractmethod
    def _reset(self):
        """
        Reset the model.
        """

    @property
    def count(self) -> NonNegativeInt:
        """
        The total amount of successes and failures collected.
        """
        return self.n_successes + self.n_failures

    @property
    def mean(self) -> Float01:
        """
        The success rate i.e. n_successes / (n_successes + n_failures).
        """
        return self.n_successes / self.count


class BaseModelMO(BaseModel, ABC):
    """
    Class to model the prior distributions of standard actions and quantitative actions for multi-objective.

    Parameters
    ----------
    models : List[BaseModelSO]
        The list of models for each objective.
    """

    models: List[BaseModelSO]

    def sample_proba(self, **kwargs) -> Union[List[MOProbability], List[QuantitativeMOProbability]]:
        """
        Sample the probability of getting a positive reward.
        """
        return [list(p) for p in zip(*[model.sample_proba(**kwargs) for model in self.models])]

    def update(self, rewards: List[List[BinaryReward]], **kwargs):
        """
        Update the model parameters.

        Parameters
        ----------
        rewards : List[List[BinaryReward]],
            if nested list, len() should follow shape of (n_samples, n_objectives)
            The binary rewards for each sample.
                If strategy is not MultiObjectiveBandit, rewards should be a list, e.g.
                    rewards = [1, 0, 1, 1, 1, ...]
                If strategy is MultiObjectiveBandit, rewards should be a list of list, e.g. (with n_objectives=2):
                    rewards = [[1, 1], [1, 0], [1, 1], [1, 0], [1, 1], ...]
        """
        if any(len(x) != len(self.models) for x in rewards):
            raise AttributeError("The shape of rewards is incorrect")

        for i, model in enumerate(self.models):
            model.update([r[i] for r in rewards], **kwargs)

    def reset(self):
        """
        Reset the model.
        """
        for model in self.models:
            model.reset()


class BaseModelCC(PyBanditsBaseModel, ABC):
    """
    Class to model action cost.

    Parameters
    ----------
    cost: Union[NonNegativeFloat, Callable[[Union[float, NonNegativeFloat]], NonNegativeFloat]]
        Cost associated to the Beta distribution.
    """

    cost: Union[NonNegativeFloat, Callable[[Union[float, np.ndarray]], NonNegativeFloat]]
